"""
    Adapter for the Flask web framework.

    This class extends the WebFrameworkAdapter and provides concrete implementations for the abstract methods.

    Methods:
        - __init__: Initialize the FlaskAdapter instance.
        - predict: Implementing the method for making predictions in Flask.
        - process_request: Implementing the method for processing requests in Flask.
        - format_response: Implementing the method for formatting responses in Flask.
        - load_model: Implementing the method for loading user BYOC code in Flask.
        - load_assets: Implementing the method to load assets in Flask.
        - start_server: Implementing the method to start the Flask server and spin up the inference endpoint.
"""

import importlib
from gru.ml_serving.ml_serving_config_reader import MLServingConfigReader
from gru.ml_serving.request_handler import RequestHandler
from gru.ml_serving.response_handler import ResponseHandler
from .web_framework_adapter import WebFrameworkAdapter

class FlaskAdapter(WebFrameworkAdapter):

    def __init__(self, config: MLServingConfigReader):
        """
        Initialize the FlaskAdapter instance.

        Parameters:
            - config: MLServingConfigReader instance for reading configuration settings.
        """
        super().__init__(config)

    def predict(self, input_data):
        """
        Make predictions in Flask.

        Parameters:
            - input_data: Input data for making predictions.

        Returns:
            Any: Predicted results in Flask.
        """
        pass

    def process_request(self, request):
        """
        Process incoming requests in Flask.

        Parameters:
            - request: Request object received from the Flask web framework.

        Returns:
            dict: Processed request data in Flask.
        """
        return RequestHandler.parse_request(request)

    def format_response(self, response):
        """
        Format and send responses in Flask.

        Parameters:
            - response_data: Data to be included in the response.

        Returns:
            Any: Formatted response in Flask.
        """
        return ResponseHandler.format_response(response)

    def load_application(self):
        """
        Load user BYOC code in Flask.

        Returns:
            Any: Loaded user code in Flask.
        """
        try:
            spec = importlib.util.spec_from_file_location("flask_app", self.config.get_config('entrypoint'))

            if spec is None:
                raise Exception(f"Error importing file: {self.config.get_config('entrypoint')}")
            
            app_module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(app_module)
            return app_module.application
        
        except Exception as e:
            raise Exception(f"Error loading flask application: {str(e)}")
        



from typing import Any, Dict, Optional
import uuid
import os

from gru.agents.framework_wrappers.memory.base import BaseMemory
from gru.agents.framework_wrappers.memory.collection_manager import CollectionManager
from gru.agents.tools.core.embeddings.embedding_factory import EmbeddingFactory, EmbeddingType
from datetime import datetime
from gru.agents.tools.core.vector_db.base import (
    AddToCollectionResponse,
    DeleteFromCollectionResponse,
    SimilaritySearchResponse,
    UpdateCollectionResponse
)
from gru.agents.tools.core.vector_db.vectordb_factory import VectorDBFactory, VectorDBType

class CansoMemory(BaseMemory):
    DEFAULT_CONFIG_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "memory_collections.yaml")


    def __init__(
        self,
        client: Any,
        embedding_type: EmbeddingType = EmbeddingType.OPENAI,
        vector_db_type: VectorDBType = VectorDBType.MILVUS,
        embedding_model: Optional[str] = None,
        config_path: Optional[str] = None,
    ):
        self.vdb_client = VectorDBFactory.get_vector_db_client(
            vector_db_type 
        )
        self.embedding_generator = EmbeddingFactory.get_embedding_generator(
            embedding_type=embedding_type,
            client=client,
            model=embedding_model
        )
        self.collection_schemas = {}

        self.config_path = config_path or self.DEFAULT_CONFIG_PATH
        
        self.collection_manager = CollectionManager(
            vector_db_client=self.vdb_client,
            embedding_dimension=self.embedding_generator.dimension,
            config_path=self.config_path
        )
        self.collection_manager.initialize_collections()

    async def store(self, data: Dict[str, Any], collection_name: str) -> AddToCollectionResponse:
        text = data.get("text", "")
        if not text:
            return AddToCollectionResponse(
                status="error",
                message="Text field is required for embedding generation",
                ids=[]
            )
        
        try:
            embedding = await self.embedding_generator.generate(text)
        except Exception as e:
            return AddToCollectionResponse(
                status="error",
                message=f"Failed to generate embedding: {str(e)}",
                ids=[]
            )
        
        
        document = data.get("data", {}).copy()
        document["text"] = text
        
        if "memory_id" not in document:
            document["memory_id"] = str(uuid.uuid4())
        
        if "created_at" not in document:
            document["created_at"] = datetime.now().isoformat()
        
        try:
            result = await self.vdb_client.add_to_collection(
                collection_name,
                embeddings=[embedding],
                data=[document]
            )
        except Exception as e:
            return AddToCollectionResponse(
                status="error",
                message=f"Failed to add document to collection: {str(e)}",
                ids=[]
            )
        
        return AddToCollectionResponse(
            status="success" if result.status != "error" else "error",
            message=result.message,
            ids=result.ids
        )

    async def retrieve(
            self, 
            query: str, 
            collection_name: str,
            top_k: int = 5
        ) -> SimilaritySearchResponse:
        try:
            query_embedding = await self.embedding_generator.generate(query)
            results = await self.vdb_client.similarity_search(
                collection_name=collection_name,
                query_vector=query_embedding,
                top_k=top_k,
            )
            return results
        except Exception as e:
            print(f"Error in retrieve: {str(e)}")
            return SimilaritySearchResponse(
                status="error",
                message=f"Failed to retrieve documents: {str(e)}",
                results=[]
            )

    async def update(
            self,
            doc_id: str,
            data: Dict[str, Any],
            collection_name: str
    ) -> UpdateCollectionResponse:
        """Update an existing memory entry in specified collection"""
        try:
            text = data.get("text")
            if text is None:
                raise ValueError("Text content is required for update")

            embedding = await self.embedding_generator.generate(text)

            # Prepare metadata with all necessary fields
            doc = {
                "text": text,
                "data": data.get("data", {}),
                "created_at": datetime.now().isoformat()
            }

            result = await self.vdb_client.update_collection(
                collection_name=collection_name,
                memory_id=doc_id,
                embedding=embedding,
                data=doc
            )

            return result
        except Exception as e:
            return UpdateCollectionResponse(
                status="error",
                message=f"Failed to update document: {str(e)}",
                id=doc_id,
                updated_count=None
            )

    async def delete(
            self, 
            doc_id: str,
            collection_name: str
        ) -> DeleteFromCollectionResponse:
        """Delete a memory entry from specified collection"""
        return await self.vdb_client.delete_from_collection(
            collection_name=collection_name,
            memory_id=doc_id
        )
        



from datetime import datetime
from typing import Dict, List, Optional, Union, Any
from gru.features.ml_feature import MLFeature
from gru.features.feature_logic import FeatureLogic
from gru.features.custom_feature_logic import CustomFeatureLogic
from gru.utils.data_types import DataType
from gru.utils.entity_type import EntityType
from gru.utils.processing_engine import ProcessingEngine
from gru.features.processing_engine_configs import StreamingProcessingEngineConfigs


class StreamingFeature(MLFeature):
    """
    A streaming raw feature is a computed column, calculated on top of a field by performing an aggregation funciton for a given timeperiod of data.

    Parameters
    ----------
    self: str
        Python self object
    name: str
        name of the feature
    description: str
        description of the feature definition
    source_name: List[str]
        data source name on which feature will be calculated
    sink_name List[str]:
        The name of the destination where the feature will be stored.
    owners: List[str]
        list of owner names
    online: bool
        If True, then feature will be persisted to Online Feature Store. By default, False
    offline: bool
        If True, then feature will be persisted to Offline Feature Store. By default, True
    feature_logic: FeatureLogic object
        FeatureLogic object that will be the logic part of feature definition

    Returns
    -------
    None

    Notes
    -----
    Use the class to create and deploy Raw features for streaming sources.

    Examples
    --------
    >>> streaming_feature_obj = StreamingFeature(
    ...     name="test_kafka_streaming_f1_v1.0.2",
    ...     description="Sample ML Feature on a Kafka Topic",
    ...     source_name=["test_kafka_streaming_source_v1.0.2"],
    ...     sink_name="demosink",
    ...     owners=["platform@yugen.ai"],
    ...     feature_logic=FeatureLogic(
    ...         field=["salary"],
    ...         time_window="3d",
    ...         timestamp_field="timestamp",
    ...         timestamp_format="yyyy-MM-dd HH:mm:ssXXX",
    ...         transform=SlidingWindowAggregation(
    ...             function="sum",
    ...             partition_by="salary",
    ...             order_by="id",
    ...             window_period="5 seconds",
    ...             window_duration="5 seconds",
    ...         ),
    ...         groupby_keys=["id", "salary"],
    ...     ),
    ...     online=True,
    ...     offline=True,
    ... )
    """

    def __init__(
        self,
        name: str,
        data_type: DataType,
        data_sources: List[str],
        offline: bool,
        feature_logic: Union[FeatureLogic,CustomFeatureLogic],
        read_option_configs: Optional[List[Dict[str, Any]]] = None,
        staging_sink_write_option_configs: Optional[Dict[str, Any]] = None,
        online_sink_write_option_configs: Optional[Dict[str, Any]] = None,
        entity: List[str] = [],
        online: bool = False,
        staging_sink: List[str] = [],
        online_sink: List[str] = [],
        description: str = "",
        owners: List[str] = [],
        k8s_configs: Dict = {},
        processing_engine: Optional[ProcessingEngine] = ProcessingEngine.PYSPARK_K8S,
        processing_engine_configs: Optional[StreamingProcessingEngineConfigs] = None,
        creation_time: datetime = datetime.now(),
        start_time: datetime = datetime.now(),
    ):
        sink = {"staging": staging_sink, "online": online_sink}
        super().__init__(
            name=name,
            type=EntityType.STREAMING_FEATURE.value,
            data_type=data_type,
            online=online,
            offline=offline,
            entity=entity,
            data_sources=data_sources,
            sink=sink,
            feature_logic=feature_logic,
            description=description,
            owners=owners,
            k8s_configs=k8s_configs,
            processing_engine=processing_engine,
            processing_engine_configs=processing_engine_configs
            or StreamingProcessingEngineConfigs(),
            creation_time=creation_time,
            start_time=start_time,
            read_option_configs=read_option_configs,
            staging_sink_write_option_configs=staging_sink_write_option_configs,
            online_sink_write_option_configs=online_sink_write_option_configs,
        )

    def to_json(self) -> Dict:
        """
        Returns a json-formatted object representing this entity.
        """
        return {
            "name": self.name,
            "type": self.type,
            "data_type": self.data_type.value,
            "data_sources": self.data_sources,
            "owners": self.owners,
            "description": self.description,
            "start_time": str(self.start_time)
            if self.start_time
            else str(datetime.now()),
            "online": self.online,
            "offline": self.offline,
            "entity": self.entity,
            "feature_logic": self.feature_logic.to_json(),
            "k8s_configs": self.k8s_configs,
            "processing_engine": self.processing_engine.value,
            "processing_engine_configs": self.processing_engine_configs.to_json(),
            "sink": self.sink,
            "creation_time": str(self.creation_time),
            "airflow_configs": self.airflow_configs,
            "created_at": str(self.creation_time),  # For backward compatibility
            "metadata": self.tags,
            "read_option_configs": self.read_option_configs,
            "staging_sink_write_option_configs": self.staging_sink_write_option_configs,
            "online_sink_write_option_configs": self.online_sink_write_option_configs,
        }

    def to_register_json(self) -> Dict:
        return self.to_json()



## Placeholders
CONTEXT_RETRIEVER_SYSTEM_PROMPT = """
You are a database expert focused on understanding context and relationships between different
database objects. Your role is to analyze queries and identify relevant context from the database.
"""

CONTEXT_RETRIEVER_USER_PROMPT_TEMPLATE = """
Find relevant context for this query:
Query: {query}

Available Context Types:
- Table schemas
- Example queries
- Documentation
- Business rules

Return the most relevant pieces of context that would help in writing an accurate SQL query.
"""



from typing import Optional
from langchain_core.tools import BaseTool
from pydantic import BaseModel, Field
from langchain_core.callbacks import CallbackManagerForToolRun
from gru.agents.framework_wrappers.memory import CansoMemory


class MemoryQueryInput(BaseModel):
    """Input for memory query operations."""
    collection_name: str = Field(
        description="Name of the memory collection to search within"
    )
    query: str = Field(
        description="Natural language query to find relevant information in memory"
    )
    top_k: int = Field(
        default=5, 
        description="Number of most relevant results to return"
    )


class MemoryRetrievalTool(BaseTool):
    """Tool for retrieving information from CansoMemory."""
    
    name: str = "query_memory"
    description: str = "Use this tool to retrieve information from memory based on natural language queries"
    args_schema: type[BaseModel] = MemoryQueryInput
    return_direct: bool = False
    memory: CansoMemory = Field(description="CansoMemory instance to use for retrieval")

    
    def __init__(self, memory: CansoMemory):
        super().__init__(memory=memory)
        self.memory = memory
    
    def _run(self, *args, **kwargs):
        raise NotImplementedError("This tool only supports async")
    
    async def _arun(
        self,
        collection_name: str,
        query: str,
        top_k: int = 5,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> str:
        try:
            results = await self.memory.retrieve(
                query=query,
                collection_name=collection_name,
                top_k=top_k
            )
            
            if not results:
                return "No relevant information found in memory."
            
            # Format the results into a readable response
            response = f"Found {len(results)} relevant memories:\n\n"
            
            for i, result in enumerate(results, 1):
                score = result.get("score", 0.0)
                text = result.get("text", "")
                
                # Format any metadata worth displaying
                metadata_str = ""
                if metadata := result.get("metadata"):
                    if isinstance(metadata, dict):
                        important_meta = {k: v for k, v in metadata.items() 
                                          if k not in ["embedding", "created_at"]}
                        if important_meta:
                            metadata_str = "\nMetadata: " + ", ".join(f"{k}: {v}" 
                                          for k, v in important_meta.items())
                
                response += f"{i}. [Relevance: {score:.2f}] {text}{metadata_str}\n\n"
            
            return response
            
        except Exception as e:
            return f"Error retrieving from memory: {str(e)}"



from typing import Optional
from langchain_core.tools import BaseTool
from pydantic import BaseModel, Field
from langchain_core.callbacks import CallbackManagerForToolRun
import boto3
from io import StringIO
import pandas as pd

def data_explainer_report(df):
    """
    Generates a data explainer report for a Pandas DataFrame.
    Args:
        df (pd.DataFrame): The DataFrame to analyze.
    Returns:
        dict: A dictionary containing the data quality report.
    """
    report = {}
    report["describe_stats"] = df.describe().to_dict()
    report["missing_values"] = df.isnull().sum().to_dict()
    report["duplicates"] = df.duplicated().sum()
    report["counts"] = {
        "rows": len(df),
        "columns": len(df.columns)
    }
    report["basics"] = {
        "columns": list(df.columns)
    }
    return report


class DataQualityInput(BaseModel):
    """Input for data quality check operations."""
    bucket: str = Field(
        description="S3 bucket in which the data is stored"
    )
    file_path: str = Field(
        description="Path of the file in the bucket"
    )


class DataQualityTool(BaseTool):
    """Tool for checking data quality of files in S3."""
    
    name: str = "check_data_quality"
    description: str = "Use this for data quality check on files stored in S3"
    args_schema: type[BaseModel] = DataQualityInput
    return_direct: bool = False
    
    def _run(self, bucket: str, file_path: str, **kwargs):
        try:
            s3_client = boto3.client('s3')
            csv_obj = s3_client.get_object(Bucket=bucket, Key=file_path)
            df = pd.read_csv(StringIO(csv_obj['Body'].read().decode('utf-8')))

            report = data_explainer_report(df)

            response = []
            response.append(f"Data Quality Report for {file_path}:")
            response.append(f"- Total rows: {report['counts']['rows']}")
            response.append(f"- Total columns: {report['counts']['columns']}")
            response.append(f"- Number of duplicates: {report['duplicates']}")

            # Report missing values if any exist
            missing = {k: v for k, v in report['missing_values'].items() if v > 0}
            if missing:
                response.append("\nMissing values found in columns:")
                for col, count in missing.items():
                    response.append(f"- {col}: {count} missing values")

            return "\n".join(response)

        except Exception as e:
            return f"Error checking data quality: {str(e)}"
    
    async def _arun(
        self,
        bucket: str,
        file_path: str,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> str:
        return self._run(bucket=bucket, file_path=file_path)



from gru.risk.workflows.workflow_commands import WorkflowCommands
import fire

class CansoCommands:
    """Root command class for the Canso CLI"""
    
    def __init__(self):
        self.workflows = WorkflowCommands()

def main():
    fire.Fire(CansoCommands)



from abc import ABC, abstractmethod
from typing import List, Optional


class CodeLoader(ABC):
    """Abstract base class for code loading from different sources."""
    
    @abstractmethod
    def retrieve_code(self, identifier: str, path: Optional[str] = None) -> str:
        """
        Retrieve code from a source.
        
        Args:
            identifier: Source identifier (e.g. repository name)
            path: Optional path within the source
            
        Returns:
            The source code as a string
        """
        pass
    
    @abstractmethod
    def list_contents(self, identifier: str, path: Optional[str] = None) -> List[str]:
        """
        List all file paths in a source.
        
        Args:
            identifier: Source identifier (e.g. repository name)
            path: Optional path within the source to start listing from
            
        Returns:
            List of file paths
        """
        pass



from abc import ABC, abstractmethod
from typing import Any, Dict


class ReadOptions(ABC):
    @abstractmethod
    def to_json(self) -> Dict[str, Any]:
        pass


class KafkaReadOptions(ReadOptions):
    def __init__(
        self,
        watermark_delay_threshold=None,
        processing_time=None,
        starting_timestamp=None,
        starting_offsets_by_timestamp=None,
        starting_offsets=None,
        fail_on_data_loss=False,
        include_headers=None,
        infer_schema=True,
        delimiter="|",
    ):
        self.watermark_delay_threshold = watermark_delay_threshold
        self.processing_time = processing_time
        self.starting_timestamp = starting_timestamp
        self.starting_offsets_by_timestamp = starting_offsets_by_timestamp
        self.starting_offsets = starting_offsets
        self.fail_on_data_loss = fail_on_data_loss
        self.include_headers = include_headers
        self.infer_schema = infer_schema
        self.delimiter = delimiter

    def to_json(self) -> Dict[str, Any]:
        return {
            "watermark_delay_threshold": self.watermark_delay_threshold,
            "processing_time": self.processing_time,
            "starting_timestamp": self.starting_timestamp,
            "starting_offsets_by_timestamp": self.starting_offsets_by_timestamp,
            "starting_offsets": self.starting_offsets,
            "fail_on_data_loss": self.fail_on_data_loss,
            "include_headers": self.include_headers,
            "infer_schema": self.infer_schema,
            "delimiter": self.delimiter,
        }



from gru.agents.tools.core.query_analyzer.base import QueryAnalyzer
from gru.agents.tools.core.code_generator.models import QueryIntent


class NLQueryAnalyzer(QueryAnalyzer):

    async def analyze_query(self, query: str) -> QueryIntent:
        pass



"""
This script defines a command-line interface (CLI) to generate standardised folder structures for ML problem statements.
The design is inspired from Cookiecutter and has been modified to adapt to Yugen's ML deployment stack.
"""
import click
import subprocess

EXPECTED_TEMPLATE_URL = "https://github.com/Yugen-ai/cookiecutter-mlops-templates.git"


@click.command()
@click.option(
    "--template",
    default="https://github.com/Yugen-ai/cookiecutter-mlops-templates.git",
    help="Cookiecutter template URL",
)
@click.option(
    "--output-dir", default=".", help="Output directory for the generated project"
)
def mlops_template_cli(template, output_dir):
    """
    Run Cookiecutter with the specified template.

    Args:
        template (str): Cookiecutter template URL.
        output_dir (str): Output directory for the generated project.
    """
    if template != EXPECTED_TEMPLATE_URL:
        click.echo(f'Error: Only the template "{EXPECTED_TEMPLATE_URL}" is supported.')
        return

    click.echo(f"Running Cookiecutter with template: {template}")
    click.echo(f"Output directory: {output_dir}")

    try:
        subprocess.run(
            ["cookiecutter", template, "--output-dir", output_dir], check=True
        )
        click.echo("Cookiecutter completed successfully!")
    except subprocess.CalledProcessError as e:
        click.echo(f"Error running Cookiecutter: {e}")


if __name__ == "__main__":
    mlops_template_cli()



from typing import Optional, List
from gru.agents.framework_wrappers import BaseAgent
from langgraph.graph import StateGraph

from gru.agents.framework_wrappers.langgraph.workflow import LanggraphWorkflow
from gru.agents.service.app import App
from gru.agents.framework_wrappers.memory import CansoMemory
import logging
from gru.agents.clients.rabbitmq import RabbitMQPublisher
from gru.agents.utils.logging import RabbitMQLogHandler


class CansoLanggraphAgent(BaseAgent):
    def __init__(
        self, 
        stateGraph: StateGraph, 
        memory: Optional[CansoMemory] = None,
        interrupt_before: Optional[List[str]] = None,
    ) -> None:
        workflow = LanggraphWorkflow(stateGraph, interrupt_before)
        self.app = App(workflow, memory)
    
    def run(self):
        self._setup_logging()
        self.app.run()
    
    def _setup_logging(self):
        try:
            rabbitmq_publisher = RabbitMQPublisher()
            rabbitmq_publisher.start()

            handler = RabbitMQLogHandler(rabbitmq_publisher)
            logging.getLogger().addHandler(handler)
        except Exception as e:
            print(f"Exception while initializing logging: {e}")



"""
Server Configuration Reader

This module provides a class, MLServingConfigReader, for reading and accessing
configuration settings from a YAML file.

Usage:
    - Instantiate MLServingConfigReader with the path to the configuration YAML file.
    - Use the provided methods to access specific configuration settings.
"""
import yaml

class MLServingConfigReader:

    def __init__(self, config_file_path):
        """
        Initialize the MLServingConfigReader instance.

        Parameters:
            - config_file_path: Path to the configuration YAML file.
        """
        self.config_file_path = config_file_path
        self.read_config()

    def read_config(self):
        """
        Read and load the configuration settings from the YAML file.
        """
        with open(self.config_file_path, 'r') as file:
            self.config = yaml.safe_load(file)

        config_validator = MLServingConfigValidator(self.config)
        config_validator.validate_config()

    def get_config(self, key):
        """
        Get the value of a specific configuration setting by key.

        Parameters:
            - key: Key for the configuration setting.

        Returns:
            Any: Value of the specified configuration setting.
        """
        return self.config.get(key)


class MLServingConfigValidator:
    """
    Configuration validator for ML serving.

    Methods:
        - __init__: Initialize the MLServingConfigValidator instance with a configuration dictionary.
        - validate_config: Validate the configuration dictionary for required keys.
        - validate_webserver_config: Validate the 'webserver' configuration for required keys.
    """

    def __init__(self, config):
        """
        Initialize the MLServingConfigValidator instance.

        Parameters:
            - config: Configuration dictionary.
        """
        self.config = config

    def validate_config(self):
        """
        Validate the configuration dictionary for required keys.
        """
        required_keys = ['ml_project', 'ml_service_name', 'serving_layer', 'entrypoint', 'webserver']
        self.check_required_key(required_keys, self.config)
            
        supported_keys = ['ml_project', 'ml_service_name', 'serving_layer', 'entrypoint', 'webserver', 'env_vars', 'models', 'artifacts']
        self.check_supported_key(supported_keys, self.config)

        self.validate_webserver_config()

    def validate_webserver_config(self):
        """
        Validate the 'webserver' configuration for required keys.
        """
        webserver_config = self.config.get('webserver')

        required_keys = ['type', 'config']
        self.check_required_key(required_keys, webserver_config)
            
        supported_keys = ['type', 'config']
        self.check_supported_key(supported_keys, webserver_config)
            
    def check_required_key(self, required_keys, config: dict):
        """
        Check if all required keys are present in the configuration.

        Raises:
            ValueError: If any required key is missing in the configuration.
        """
        for key in required_keys:
            if key not in config:
                raise ValueError(f"Missing required key: {key} in config yaml")
    
    def check_supported_key(self, supported_keys, config):
        """
        Check if all keys in the configuration are supported.

        Raises:
            ValueError: If any key in the configuration is not supported.
        """
        for key in config:
            if key not in supported_keys:
                raise ValueError(f"Unsupported key: {key} in config yaml")



import pytest
from unittest.mock import patch, MagicMock
from gru.features.processing_engine_configs import ProcessingEngineConfigs, StreamingProcessingEngineConfigs, BatchProcessingEngineConfigs

@pytest.fixture
def mock_config_reader():
    with patch('gru.features.processing_engine_configs.ConfigReader') as mock:
        config = MagicMock()
        config.deploy_mode = "cluster"
        config.driver_memory = "4g"
        config.executor_memory = "2g"
        config.service_account = "spark"
        config.aws_credential_provider = "com.amazonaws.auth.WebIdentityTokenCredentialsProvider"
        config.container_image = "spark:latest"
        config.container_image_pull_secrets = "regcred"
        config.file_upload_path = "s3a://bucket/path"
        config.name_space = "default"
        config.num_executors = 2
        config.executor_cores = 2
        config.driver_cores = 1
        config.driver_limit_cores = "1200m"
        config.executor_limit_cores = "1200m"
        config.executor_instances = 2
        config.checkpoint_path = "s3a://sample-tenant-1-canso-materialised-features/ml_feature_store/checkpoints/clicks_per_display"
        mock.return_value = config
        yield mock

def test_processing_engine_configs_initialization(mock_config_reader):
    configs = ProcessingEngineConfigs(
        deploy_mode="client",
        driver_memory="8g",
        executor_memory="4g",
        service_account="test-account",
        aws_credential_provider="test-provider",
        container_image="test-image:latest",
        container_image_pull_secrets="test-secret",
        file_upload_path="s3a://test-bucket/path",
        name_space="test-namespace",
        num_executors=4,
        executor_cores=4,
        driver_cores=2,
        driver_limit_cores="2400m",
        executor_limit_cores="2400m",
        executor_instances=4,
        checkpoint_path="s3a://sample-tenant-1-canso-materialised-features/ml_feature_store/checkpoints/clicks_per_display"
    )

    assert configs.deploy_mode == "client"
    assert configs.driver_memory == "8g"
    assert configs.executor_memory == "4g"
    assert configs.service_account == "test-account"
    assert configs.aws_credential_provider == "test-provider"
    assert configs.container_image == "test-image:latest"
    assert configs.container_image_pull_secrets == "test-secret"
    assert configs.file_upload_path == "s3a://test-bucket/path"
    assert configs.name_space == "test-namespace"
    assert configs.num_executors == 4
    assert configs.executor_cores == 4
    assert configs.driver_cores == 2
    assert configs.driver_limit_cores == "2400m"
    assert configs.executor_limit_cores == "2400m"
    assert configs.executor_instances == 4
    assert configs.checkpoint_path == "s3a://sample-tenant-1-canso-materialised-features/ml_feature_store/checkpoints/clicks_per_display"

def test_processing_engine_configs_to_json(mock_config_reader):
    configs = ProcessingEngineConfigs()
    json_output = configs.to_json()

    assert json_output["deploy-mode"] == "cluster"
    assert json_output["spark.kubernetes.driver.memory"] == "4g"
    assert json_output["spark.kubernetes.executor.memory"] == "2g"
    assert json_output["spark.kubernetes.authenticate.driver.serviceAccountName"] == "spark"
    assert json_output["spark.hadoop.fs.s3a.aws.credentials.provider"] == "com.amazonaws.auth.WebIdentityTokenCredentialsProvider"
    assert json_output["spark.kubernetes.container.image"] == "spark:latest"
    assert json_output["spark.kubernetes.container.image.pullSecrets"] == "regcred"
    assert json_output["spark.kubernetes.file.upload.path"] == "s3a://bucket/path"
    assert json_output["spark.kubernetes.namespace"] == "default"
    assert json_output["num-executors"] == 2
    assert json_output["spark.executor.cores"] == 2
    assert json_output["spark.driver.cores"] == 1
    assert json_output["spark.kubernetes.driver.limit.cores"] == "1200m"
    assert json_output["spark.kubernetes.executor.limit.cores"] == "1200m"
    assert json_output["spark.executor.instances"] == 2
    assert json_output["checkpoint_path"] == "s3a://sample-tenant-1-canso-materialised-features/ml_feature_store/checkpoints/clicks_per_display"

def test_streaming_processing_engine_configs(mock_config_reader):
    configs = StreamingProcessingEngineConfigs()
    assert configs.default_config_path == "./gru/config/features/default_processing_engine_configs_streaming.yaml"
    mock_config_reader.assert_called_once_with(config_path="./gru/config/features/default_processing_engine_configs_streaming.yaml")

def test_batch_processing_engine_configs(mock_config_reader):
    configs = BatchProcessingEngineConfigs()
    assert configs.default_config_path == "./gru/config/features/default_processing_engine_configs_batch.yaml"
    mock_config_reader.assert_called_once_with(config_path="./gru/config/features/default_processing_engine_configs_batch.yaml")

def test_processing_engine_configs_override_default(mock_config_reader):
    configs = ProcessingEngineConfigs(
        deploy_mode="client",
        driver_memory="8g",
        executor_memory="4g"
    )
    json_output = configs.to_json()

    assert json_output["deploy-mode"] == "client"
    assert json_output["spark.kubernetes.driver.memory"] == "8g"
    assert json_output["spark.kubernetes.executor.memory"] == "4g"
    # Other fields should still use default values
    assert json_output["spark.kubernetes.authenticate.driver.serviceAccountName"] == "spark"

if __name__ == "__main__":
    pytest.main()


import json
import os
from typing import Any, List

import psycopg
from psycopg_pool import AsyncConnectionPool
from gru.agents.checkpoint.postgres import PostgresAsyncConnectionPool
from gru.agents.framework_wrappers import AgentWorkflow
from langgraph.graph import StateGraph

from gru.agents.schemas import AgentInvokeRequest
from langchain_core.runnables import RunnableConfig
from langgraph.checkpoint.postgres.aio import AsyncPostgresSaver

from gru.agents.schemas.schemas import (
    AgentConversationRequest,
    AgentConversationResponse,
    TaskCompleteRequest,
)
from langchain_core.messages import ToolMessage


class LanggraphWorkflow(AgentWorkflow):

    def __init__(
        self, stateGraph: StateGraph, interrupt_before: List[str] | None = None
    ) -> None:
        super().__init__()
        self.state_graph = stateGraph
        self.interrupt_before = interrupt_before

    async def setup(self):
        checkpoint_db_type = os.getenv("CHECKPOINT_DB_TYPE", "postgres")
        if checkpoint_db_type == "postgres":
            pool = PostgresAsyncConnectionPool().get()
            checkpointer = await self._setup_postgres_checkpointer(pool)
            if self.interrupt_before is not None:
                self.compiled_graph = self.state_graph.compile(
                    checkpointer=checkpointer, interrupt_before=self.interrupt_before
                )
            else:
                self.compiled_graph = self.state_graph.compile(
                    checkpointer=checkpointer
                )

    async def _setup_postgres_checkpointer(self, pool: AsyncConnectionPool):
        checkpointer = AsyncPostgresSaver(pool)
        async with pool.connection() as conn:
            async with conn.cursor() as cur:
                try:
                    await cur.execute(
                        """
                        SELECT EXISTS (
                            SELECT FROM information_schema.tables 
                            WHERE  table_schema = 'public'
                            AND    table_name   = 'checkpoints'
                        );
                    """
                    )
                    table_exists = (await cur.fetchone())[0]

                    if not table_exists:
                        print("Checkpoints table does not exist. Running setup...")
                        await checkpointer.setup()
                    else:
                        print("Checkpoints table already exists. Skipping setup.")
                except psycopg.Error as e:
                    print(f"Error checking for checkpoints table: {e}")
                    raise e
        return checkpointer

    async def invoke(self, request: AgentInvokeRequest) -> dict[str, Any]:
        config = RunnableConfig(
            configurable={"thread_id": request.prompt_id},
        )
        return await self.compiled_graph.ainvoke(
            input=request.prompt_body, config=config
        )

    async def resume(self, request: TaskCompleteRequest) -> dict[str, Any]:
        config = RunnableConfig(
            configurable={"thread_id": request.prompt_id},
        )
        state = await self.compiled_graph.aget_state(config=config)

        ## todo : Support for Custom state?
        await self.compiled_graph.aupdate_state(
            config=config,
            values={
                "messages": [
                    ToolMessage(
                        content=json.dumps(request.result),
                        name=request.task_type,
                        tool_call_id=request.tool_call_id,
                    )
                ]
            },
            as_node=state.next[0],
        )

        result = await self.compiled_graph.ainvoke(input=None, config=config)
        return result

    async def converse(
        self, request: AgentConversationRequest
    ) -> AgentConversationResponse:
        config = RunnableConfig(configurable={"thread_id": request.conversation_id})
        user_message = request.message
        messages = await self._get_invocation_message_based_on_state(
            config, user_message
        )

        await self.compiled_graph.ainvoke(messages, config=config)
        agent_message = await self._get_agent_message_based_on_state(config)

        return AgentConversationResponse(message=agent_message)

    async def _get_invocation_message_based_on_state(self, config, user_message):
        snapshot = await self.compiled_graph.aget_state(config)
        pending_tool_calls = self._get_pending_tool_calls(snapshot)
        if pending_tool_calls and len(pending_tool_calls) > 0:
            # Scenario 1: when graph was interrupted to get toolcall confirmation from user.
            if user_message.strip() == "y" or user_message.strip() == "Y":
                # Scenario 1.1: when user confirms the tool call. The graph continues with tool call execution.
                messages = None
            else:
                # Scenario 1.2: when user denies tool call with some input. Tool response is fabricated as below.
                messages = {
                    "messages": [
                        ToolMessage(
                            tool_call_id=pending_tool_calls[0]["id"],
                            content=f"Tool call denied by user. Reasoning: '{user_message}'. Continue assisting, accounting for the user's input.",
                        )
                    ]
                }
        else:
            # Scenario 2: Normal user message
            messages = {"messages": ("user", user_message)}
        return messages

    async def _get_agent_message_based_on_state(self, config):
        snapshot = await self.compiled_graph.aget_state(config)
        pending_tool_calls = self._get_pending_tool_calls(snapshot)
        if pending_tool_calls and len(pending_tool_calls) > 0:
            # Scenario 1: When graph is interrupted to get tool call confirmation from user
            result = self._prepare_tool_call_confirmation_message(pending_tool_calls)
        else:
            # Scenario 2: Normal Agent Message
            result = snapshot.values["messages"][-1].content
        return result

    def _prepare_tool_call_confirmation_message(self, pending_tool_calls):
        result = f"I'll be calling the tool {pending_tool_calls[0]['name']} with following arguments:\n"
        args = pending_tool_calls[0]["args"]
        for arg in args:
            result = result + f"{arg}: {args[arg]}\n"
        result = (
            result
            + "Do you approve of this action? Type 'y' to continue; otherwise, explain your requested changed."
        )
        return result

    def _get_pending_tool_calls(self, snapshot):
        return snapshot.values["messages"][-1].tool_calls if snapshot.next else None



from enum import Enum
from gru.agents.tools.core.vector_db.milvus import MilvusClient
from gru.agents.tools.core.vector_db.base import VectorDBClient

class VectorDBType(Enum):
    """Enum for supported vector database types"""
    MILVUS = "milvus"

class VectorDBFactory:
    """Factory class for creating vector database clients"""
    @staticmethod
    def get_vector_db_client(
        vector_db_type: VectorDBType, 
    ) -> VectorDBClient:
        """
        Create a vector database client based on the specified type.
        
        Args:
            vector_db_type: Type of vector database to create
            **kwargs: Additional arguments to pass to the constructor
            
        Returns:
            An instance of the requested VectorDBClient
        """
        if vector_db_type == VectorDBType.MILVUS:
            return MilvusClient()
        else:
            raise ValueError(f"Unsupported vector database type: {vector_db_type}")



"""
Defines a S3 Data Source
"""

from datetime import datetime
from typing import Any, Optional, Union, Dict
from gru.sources.datasource import DataSource
from gru.sources.file_types import CSVType, ParquetType
from gru.utils.data_processing_paradigms import DataProcessingParadigm
from gru.utils.cloud_providers import CloudProviders


class S3DataSource(DataSource):
    """
    Represents an AWS S3 Data Source.

    Parameters:
    ----------
    name (str):
        Name of the S3 data source object.
    bucket (str):
        S3 bucket name.
    base_key (str):
        Fixed data path inside the bucket.
    varying_key_suffix_format (str):
        Varying key format of data based on loading date in S3 (default is "d=%Y-%m-%d/h=%H%M").
    varying_key_suffix_freq (str):
        Frequency of data loading into the S3 object.
    description (str):
        Description of the S3 data source object.
    owners (Union[List[str], str]):
        List of owners of the data source.
    event_timestamp_field (str):
        Timestamp field in the data.
    file_type (Union[CSVType, ParquetType]):
        Type of the file (e.g., CSV, PARQUET).
    schema (Schema):
        Object of the `Schema` class defining the data schema in S3.
    time_offset (int):
        Time offset from 00:00:00 in minutes (default is 0, e.g., 60 means data starts at 00:01:00).
    event_timestamp_format (str):
        Format of the timestamp field in the data source (default is "%Y-%m-%d %H:%M").
    created_at (datetime):
        Timestamp when the data source was registered (default is the current timestamp).

    Notes:
    ------
    This class represents an S3 data source for batch data processing.

    Examples:
    ---------
    >>> data_source_obj = S3DataSource(
    ...     name="my_data_source_11",
    ...     bucket="market-intelligence-platform",
    ...     base_key="sales/raw_data",
    ...     varying_key_suffix_format="d=%Y-%m-%d/h=%H%M",
    ...     varying_key_suffix_freq="15min",
    ...     time_offset=5*60,
    ...     description="random desc of data source",
    ...     owners=["xyz"],
    ...     created_at=datetime(2022, 12, 20, 00, 00, 00),
    ...     file_type=CSVType(header=True),
    ...     schema=schema_obj,
    ...     event_timestamp_field="viewed_at",
    ...     event_timestamp_format="%Y-%m-%d %H:%M",
    ... )
    """

    def __init__(
        self,
        name,
        bucket,
        base_key,
        varying_key_suffix_format,
        varying_key_suffix_freq,
        description,
        owners,
        event_timestamp_field,
        file_type: Union[CSVType, ParquetType],
        schema,
        time_offset=0,
        metadata: Optional[Dict[str, str]] = {},
        additional_read_configs: Optional[Dict[str, Any]] = {},
        event_timestamp_format="%Y-%m-%d %H:%M",
        created_at=datetime.now(),
    ):
        physical_uri = {
            "bucket": bucket,
            "base_key": base_key,
            "varying_key_suffix_format": varying_key_suffix_format,
            "varying_key_suffix_freq": varying_key_suffix_freq,
            "time_offset": time_offset,
            "uri_scheme": "s3",
            "uri_scheme_pyspark": "s3a"
        }
        super().__init__(
            name=name,
            description=description,
            family="object_storage",
            processing_paradigm=DataProcessingParadigm.Batch,
            cloud_provider=CloudProviders.AWS,
            additional_read_configs=additional_read_configs,
            physical_uri=physical_uri,
            schema=schema,
            file_type=file_type,
            type="S3DataSource",
            event_timestamp_field=event_timestamp_field,
            event_timestamp_format=event_timestamp_format,
            metadata=metadata,
            owners=owners,
            created_at=created_at,
        )

    def to_json(self) -> Dict:
        """
        Convert the S3DataSource instance to a JSON-compatible dictionary.

        Returns:
        Dict: A dictionary representing the S3DataSource.
        """
        return {
            "name": self.name,
            "description": self.description,
            "family": self.family,
            "processing_paradigm": self.processing_paradigm.value,
            "cloud_provider": self.cloud_provider.value
            if self.cloud_provider
            else None,
            "additional_read_configs": self.additional_read_configs,
            "physical_uri": self.physical_uri,
            "file_type": self.file_type.type if self.file_type else None,
            "file_type_properties": self.file_type.options
            if self.file_type.options
            else {},
            "schema": self.schema.schema,
            "type": self.type,
            "event_timestamp_field": self.event_timestamp_field,
            "event_timestamp_format": self.event_timestamp_format,
            "metadata": self.metadata,
            "owners": self.owners,
            "created_at": str(self.created_at),
        }



from typing import Optional
from pydantic import BaseModel, Field, validator

class HPAConfig(BaseModel):
    """Schema for HPA configuration"""
    min_replicas: int = Field(1, ge=1, description="Minimum number of replicas")
    max_replicas: int = Field(..., ge=1, description="Maximum number of replicas")
    target_cpu_utilization_percentage: Optional[int] = Field(
        None, ge=1, le=100,
        description="Target CPU utilization percentage"
    )
    target_memory_utilization_percentage: Optional[int] = Field(
        None, ge=1, le=100,
        description="Target memory utilization percentage"
    )

    @validator('max_replicas')
    def validate_max_replicas(cls, v, values):
        if 'min_replicas' in values and v < values['min_replicas']:
            raise ValueError("max_replicas must be greater than or equal to min_replicas")
        return v

    class Config:
        allow_population_by_field_name = True



from pydantic import BaseModel, Field
from typing import Optional, List, Dict, Any


class TextToSQLToolInput(BaseModel):
    query: str = Field(description="Natural language query to convert to SQL")
    table_info: Optional[List[Dict[str, Any]]] = Field(
        default=None,
        description="Optional table information to initialize or update the context",
    )



import os
import json
import pytest
import yaml
from unittest.mock import patch, mock_open

from gru.cli import GruCommands, AgentCommands, MemoryCommands, ComponentCommands
from gru.utils.constants import TOKEN_ENV_VAR_NAME

TEST_TOKEN = "test_auth_token"
TEST_CORRELATION_ID = "test-correlation-id"


@pytest.fixture
def mock_uuid():
    with patch("uuid.uuid4") as mock:
        mock.return_value = TEST_CORRELATION_ID
        yield mock


@pytest.fixture
def gru_commands():
    return GruCommands()


@pytest.fixture
def agent_commands():
    return AgentCommands(auth_token=TEST_TOKEN)


@pytest.fixture
def memory_commands():
    return MemoryCommands(auth_token=TEST_TOKEN)


@pytest.fixture
def component_commands():
    return ComponentCommands(auth_token=TEST_TOKEN)


@pytest.fixture
def mock_config_file():
    config_data = {
        "agent_name": "Test Agent",
        "task_server_name": "test-task-server",
        "checkpoint_db_name": "test-checkpoint-db",
        "replicas": 1,
        "iam_role_arn": "test-iam-role",
        "vector_db_name": "test-vector-db",
    }

    with patch(
        "builtins.open", mock_open(read_data=yaml.dump(config_data))
    ) as mock_file:
        yield mock_file


@pytest.fixture
def mock_memory_file():
    memory_data = {
        "collection_name": "test_collection",
        "text": "Test memory text",
        "data": {"key": "value"},
        "memory_id": "test-memory-id",
    }

    with patch(
        "builtins.open", mock_open(read_data=json.dumps(memory_data))
    ) as mock_file:
        yield mock_file


class TestComponentCommands:
    def test_get_token_with_provided_token(self, component_commands):
        """Test that get_token returns the provided token"""
        assert component_commands.get_token() == TEST_TOKEN

    def test_get_token_from_env(self):
        """Test that get_token falls back to environment variable"""
        with patch.dict(os.environ, {TOKEN_ENV_VAR_NAME: "env_token"}):
            component = ComponentCommands()
            assert component.get_token() == "env_token"

    @patch("gru.cli.component_setup")
    def test_setup(self, mock_setup, component_commands, mock_uuid):
        """Test component setup command"""
        mock_setup.return_value = "Component setup successful"

        result = component_commands.setup("test-cluster", "config.yaml")

        assert result == "Component setup successful"
        mock_setup.assert_called_once_with(
            TEST_CORRELATION_ID, "test-cluster", "config.yaml", TEST_TOKEN
        )

    @patch("gru.cli.component_setup")
    def test_setup_file_not_found(self, mock_setup, component_commands):
        """Test component setup with file not found error"""
        mock_setup.side_effect = FileNotFoundError()

        result = component_commands.setup("test-cluster", "not-found.yaml")

        assert "Error: not-found.yaml file not found." in result
        mock_setup.assert_called_once()


class TestAgentCommands:
    def test_get_token_with_provided_token(self, agent_commands):
        """Test that get_token returns the provided token"""
        assert agent_commands.get_token() == TEST_TOKEN

    @patch("gru.cli.ai_agent_templates_setup")
    def test_create_bootstrap(self, mock_bootstrap, agent_commands):
        """Test create bootstrap command"""
        result = agent_commands.create_bootstrap()

        assert "Agent bootstrap project created successfully" in result
        mock_bootstrap.assert_called_once()

    @patch("gru.cli.register_agent")
    def test_register(self, mock_register, agent_commands, mock_config_file, mock_uuid):
        """Test register agent command"""
        mock_register.return_value = "Agent registered successfully"

        result = agent_commands.register(
            "agent_folder", "test-cluster", "test-image", "test-pull-secret"
        )

        assert result == "Agent registered successfully"
        mock_register.assert_called_once_with(
            TEST_CORRELATION_ID,
            TEST_TOKEN,
            "agent_folder",
            "test-cluster",
            "test-image",
            "test-pull-secret",
        )

    @patch("gru.cli.deploy_agent")
    def test_deploy(self, mock_deploy, agent_commands, mock_uuid):
        """Test deploy agent command"""
        mock_deploy.return_value = "Agent deployed successfully"

        result = agent_commands.deploy("test-agent")

        assert result == "Agent deployed successfully"
        mock_deploy.assert_called_once_with(
            TEST_CORRELATION_ID, TEST_TOKEN, "test-agent"
        )

    @patch("gru.cli.update_agent")
    def test_update(self, mock_update, agent_commands, mock_config_file, mock_uuid):
        """Test update agent command"""
        mock_update.return_value = "Agent updated successfully"

        result = agent_commands.update("agent_folder", "new-image", "new-pull-secret")

        assert result == "Agent updated successfully"
        mock_update.assert_called_once_with(
            TEST_CORRELATION_ID,
            TEST_TOKEN,
            "agent_folder",
            "new-image",
            "new-pull-secret",
        )

    @patch("gru.cli.delete_agent")
    def test_delete(self, mock_delete, agent_commands, mock_uuid):
        """Test delete agent command"""
        mock_delete.return_value = "Agent deleted successfully"

        result = agent_commands.delete("test-agent")

        assert result == "Agent deleted successfully"
        mock_delete.assert_called_once_with(
            TEST_CORRELATION_ID, TEST_TOKEN, "test-agent"
        )

    @patch("gru.cli.read_prompt_file")
    @patch("gru.cli.prompt_agent")
    def test_prompt(self, mock_prompt, mock_read_file, agent_commands):
        """Test prompt agent command"""
        mock_read_file.return_value = {"prompt": "Test prompt"}
        mock_prompt.return_value = "Prompt sent successfully"

        result = agent_commands.prompt("test-agent", "prompt.json")

        assert result == "Prompt sent successfully"
        mock_read_file.assert_called_once_with("prompt.json")
        mock_prompt.assert_called_once_with(
            "test-agent", {"prompt": "Test prompt"}, TEST_TOKEN
        )

    @patch("gru.cli.converse_agent")
    def test_converse(self, mock_converse, agent_commands):
        """Test converse agent command"""
        mock_converse.return_value = "Conversation Stopped"

        result = agent_commands.converse("test-agent")

        assert result == "Conversation Stopped"
        mock_converse.assert_called_once_with(TEST_TOKEN, "test-agent", None)

    @patch("gru.cli.converse_agent")
    def test_converse_with_id(self, mock_converse, agent_commands):
        """Test converse agent command with conversation ID"""
        mock_converse.return_value = "Conversation Stopped"

        result = agent_commands.converse("test-agent", "test-conversation")

        assert result == "Conversation Stopped"
        mock_converse.assert_called_once_with(
            TEST_TOKEN, "test-agent", "test-conversation"
        )


class TestMemoryCommands:
    def test_get_token_with_provided_token(self, memory_commands):
        """Test that get_token returns the provided token"""
        assert memory_commands.get_token() == TEST_TOKEN

    @patch("gru.cli.memory_insert")
    def test_insert(self, mock_insert, memory_commands, mock_memory_file, mock_uuid):
        """Test memory insert command"""
        mock_insert.return_value = "Memory inserted successfully"

        result = memory_commands.insert("test-agent", "memory_data.json")

        assert result == "Memory inserted successfully"
        mock_insert.assert_called_once_with(
            TEST_CORRELATION_ID, TEST_TOKEN, "test-agent", "memory_data.json"
        )

    @patch("gru.cli.memory_update")
    def test_update(self, mock_update, memory_commands, mock_memory_file, mock_uuid):
        """Test memory update command"""
        mock_update.return_value = "Memory updated successfully"

        result = memory_commands.update("test-agent", "memory_data.json")

        assert result == "Memory updated successfully"
        mock_update.assert_called_once_with(
            TEST_CORRELATION_ID, TEST_TOKEN, "test-agent", "memory_data.json"
        )

    @patch("gru.cli.memory_delete")
    def test_delete(self, mock_delete, memory_commands, mock_uuid):
        """Test memory delete command"""
        mock_delete.return_value = "Memory deleted successfully"

        result = memory_commands.delete(
            "test-agent", "test-memory-id", "test-collection"
        )

        assert result == "Memory deleted successfully"
        mock_delete.assert_called_once_with(
            TEST_CORRELATION_ID,
            TEST_TOKEN,
            "test-agent",
            "test-memory-id",
            "test-collection",
        )


class TestErrorHandling:
    @patch("gru.cli.deploy_agent")
    def test_deploy_value_error(self, mock_deploy, agent_commands):
        """Test that value errors are handled properly"""
        mock_deploy.side_effect = ValueError("Invalid agent name")

        result = agent_commands.deploy("invalid-agent")

        assert "Invalid agent name" in result
        mock_deploy.assert_called_once()

    @patch("gru.cli.deploy_agent")
    def test_deploy_unexpected_error(self, mock_deploy, agent_commands, mock_uuid):
        """Test that unexpected errors are handled properly"""
        mock_deploy.side_effect = Exception("Unexpected error")

        result = agent_commands.deploy("test-agent")

        assert (
            f"An unexpected error occured. Correlation ID: {TEST_CORRELATION_ID}"
            in result
        )
        mock_deploy.assert_called_once()

    @patch("gru.cli.read_prompt_file")
    def test_prompt_file_not_found(self, mock_read_file, agent_commands):
        """Test prompt with file not found error"""
        mock_read_file.side_effect = FileNotFoundError("File not found")

        result = agent_commands.prompt("test-agent", "not-found.json")

        assert "File error: File not found" in result
        mock_read_file.assert_called_once()

    @patch("gru.cli.read_prompt_file")
    def test_prompt_json_error(self, mock_read_file, agent_commands):
        """Test prompt with JSON decode error"""
        mock_read_file.side_effect = json.JSONDecodeError("Invalid JSON", "", 0)

        result = agent_commands.prompt("test-agent", "invalid.json")

        assert "JSON error" in result
        mock_read_file.assert_called_once()


class TestEnvironmentToken:
    def test_environment_token_fallback(self):
        """Test that token falls back to environment when not provided"""
        with patch.dict(os.environ, {TOKEN_ENV_VAR_NAME: "env_token"}):
            agent_cmd = AgentCommands()  # No token provided, should use environment

            assert agent_cmd.get_token() == "env_token"
            assert agent_cmd.memory.get_token() == "env_token"

    def test_missing_environment_token(self):
        """Test error when token not provided and not in environment"""
        with patch.dict(os.environ, {}, clear=True):  # Clear environment variables
            agent_cmd = AgentCommands()  # No token provided or in environment

            with pytest.raises(
                ValueError, match=f"Environment variable {TOKEN_ENV_VAR_NAME} missing"
            ):
                agent_cmd.get_token()



"""
Abstract base class representing a batch feature definition.
"""
from abc import abstractmethod
import json
from typing import Any, List, Dict, Union, Optional
from datetime import datetime
from gru.schema.registrable import Registrable
from gru.utils.data_types import DataType
from gru.features.feature_logic import FeatureLogic
from gru.features.custom_feature_logic import CustomFeatureLogic
from gru.features.processing_engine_configs import BatchProcessingEngineConfigs
from gru.utils.processing_engine import ProcessingEngine

class MLFeature(Registrable):
    """
    This abstract class defines the common attributes and methods for batch features
    that are used in data processing pipelines.

    Attributes:
        name (str): The name of the batch feature.
        type (str): The type of the feature.
        data_type (DataTypes): The data type of the feature.
        online (bool): Flag indicating if the feature is available online.
        offline (bool): Flag indicating if the feature is available offline.
        entity (List[str]): List of entities associated with the feature.
        source_name (Union[str, List[str]): The source or sources of the feature.
        sink_name (str): The destination or sink for the feature.
        feature_logic (FeatureLogic): The logic or transformation applied to the feature.
        description (str): Description or additional information about the feature.
        owners (Union[str, List[str]): List of owners responsible for the feature.
        feature_start_time (datetime): The timestamp when the feature starts being available.
        schedule (str): The scheduling interval for batch feature generation.
        active (bool): Flag indicating if the feature is active.
        k8s_configs (Dict): Kubernetes configurations for feature processing.
        creation_time (datetime): The timestamp of feature creation.

    Abstract Methods:
        to_json(self) -> dict:
        Returns a JSON representation of the batch feature.

    """

    def __init__(
        self,
        name: str,
        type: str,
        data_type: DataType,
        processing_engine: Optional[ProcessingEngine],
        processing_engine_configs: Optional[BatchProcessingEngineConfigs],
        online: bool,
        offline: bool,
        read_option_configs: Optional[List[Dict[str, Any]]],
        staging_sink_write_option_configs: Optional[Dict[str, Any]],
        online_sink_write_option_configs: Optional[Dict[str, Any]],
        entity: List[str],
        data_sources: Union[str, List[str]],
        sink: Dict,
        feature_logic: Union[FeatureLogic, CustomFeatureLogic] = None,
        description: str = "",
        owners: Union[str, List[str]] = [],
        start_time: datetime = None,
        schedule: str = "30 0 * * *",
        active: bool = False,
        k8s_configs: Dict = {},
        airflow_configs: Dict = {},
        creation_time: datetime = datetime.now(),
        tags: Dict[str, Any] = {}
    ):
        self.name = name
        self.data_type = data_type
        self.data_sources = data_sources
        self.entity = entity
        self.type = type
        self.entity_type = type
        self.owners = owners
        self.description = description
        self.start_time = start_time
        self.online = online
        self.offline = offline
        self.schedule = schedule
        self.feature_logic = feature_logic
        self.sink = sink
        self.active = active
        self.k8s_configs = k8s_configs
        self.creation_time = creation_time
        self.processing_engine = processing_engine
        self.processing_engine_configs = processing_engine_configs
        self.airflow_configs = airflow_configs
        self.tags = tags
        self.read_option_configs=read_option_configs
        self.staging_sink_write_option_configs=staging_sink_write_option_configs
        self.online_sink_write_option_configs=online_sink_write_option_configs

    @abstractmethod
    def to_json(self) -> Dict:
        """
        Returns a json-formatted object representing this entity.
        """
        pass

    def __str__(self) -> str:
        return json.dumps(self.to_json(), indent=2, sort_keys=True)

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and self.to_json() == other.to_json()

    def to_register_json(self) -> Dict:
        return self.to_json()



"""
    Gunicorn Server Wrapper.

    This class extends MLServerInterface and provides a wrapper for starting a Gunicorn server using the BaseApplication class.
    It allows easy integration of Gunicorn with a Flask application.

    Methods:
        - __init__: Initialize the GunicornServer instance.
        - start: Start the Gunicorn server.
        - stop: Stop the Gunicorn server (Not implemented).
        - reload: Reload the Gunicorn server.
        - load: Load the Flask application.
"""

import gunicorn.app.base
from .ml_server_interface import MLServerInterface

class GunicornServer(MLServerInterface):

    def __init__(self, app, options=None):
        """
        Initialize the GunicornServer instance.

        Parameters:
            - app: The Flask application instance.
            - options (dict): Optional Gunicorn configuration options.
        """
        self.gunicorn_app = GunicornBaseApplication(app, options)

    def start(self):
        """Start the Gunicorn server."""
        self.gunicorn_app.run()

    def stop(self):
        """Stop the Gunicorn server."""
        pass

    def reload(self):
        """Reload the Gunicorn server."""
        self.gunicorn_app.reload()

    def load(self):
        """Load the Flask application."""
        pass

class GunicornBaseApplication(gunicorn.app.base.BaseApplication):
    """
    Gunicorn Base Application Wrapper. see https://github.com/benoitc/gunicorn/blob/master/gunicorn/app/base.py

    This class provides a wrapper for starting a Gunicorn server using the BaseApplication class.
    It allows easy integration of Gunicorn with a Flask application.

    Methods:
        - __init__: Initialize the GunicornServer instance.
        - load_config: Load Gunicorn configuration from options.
        - load: Load the Flask application (required by the BaseApplication class).
    """

    def __init__(self, app, options=None):
        """
        Initialize the GunicornServer instance.

        Parameters:
            - app: The Flask application instance.
            - options (dict): Optional Gunicorn configuration options.
        """
        self.options = options or {}
        self.application = app
        super().__init__()

    def load_config(self):
        """Load Gunicorn configuration from options."""
        for key, value in self.options.items():
            self.cfg.set(key, value)

    def load(self):
        """Load the Flask application (required by the BaseApplication class)."""
        return self.application



from typing import List, Optional
from github import Github, Auth
from gru.agents.tools.core.code_loader.base import CodeLoader
import os

class GitHubLoader(CodeLoader):
    """Implementation of CodeLoader for GitHub repositories."""
    
    def __init__(self, token: Optional[str] = None):
        """
        Initialize with GitHub token.
        
        Args:
            token: GitHub authentication token. If None, will attempt to use GITHUB_TOKEN environment variable.
        """
        
        self.token = token or os.environ.get("GITHUB_TOKEN")
        
        if self.token is None:
            raise ValueError("GitHub token not provided and GITHUB_TOKEN environment variable not set")

    def _get_repo(self, identifier: str):
        """
        Get a GitHub repository object.
        
        Args:
            identifier: Repository name in format "owner/repo"
            
        Returns:
            A GitHub repository object
        """
        auth = Auth.Token(self.token)
        g = Github(auth=auth)
        return g.get_repo(identifier)
    
        
    def retrieve_code(self, identifier: str, path: Optional[str] = None) -> str:
        """
        Retrieve code from a GitHub repository.
        
        Args:
            identifier: Repository name in format "owner/repo"
            path: File path within the repository
            
        Returns:
            The source code as a string
        """
        
        repo = self._get_repo(identifier)
        
        contents = repo.get_contents(path)
        return contents.decoded_content.decode()
    
    def list_contents(self, identifier: str, path: Optional[str] = "") -> List[str]:
        """
        List all file paths in a GitHub repository.
        
        Args:
            identifier: Repository name in format "owner/repo"
            path: Optional directory path within the repository to start listing from
            
        Returns:
            List of file paths in the repository
        """
        
        repo = self._get_repo(identifier)
        contents = repo.get_contents(path or "")
        
        result = []
        while contents:
            file_content = contents.pop(0)
            if file_content.type == "dir":
                contents.extend(repo.get_contents(file_content.path))
            else:
                result.append(file_content.path)
                
        return result



from gru.agents.tools.core.code_generator.base import CodeGenerator
from gru.agents.tools.core.code_generator.models import (
    QueryIntent,
    RetrievalResult,
    SQLQueryResult,
)
from gru.agents.prompts.sql import (
    SQL_GENERATOR_USER_PROMPT_TEMPLATE,
    SQL_GENERATOR_SYSTEM_PROMPT,
)
import json


class SQLCodeGenerator(CodeGenerator):
    def __init__(
        self,
        llm_client,
        system_prompt=None,
        user_prompt_template=None
    ):
        """
        Initializes the SQL Code Generator with optional custom prompts.

        :param llm_client: The LLM client to use for generation.
        :param system_prompt: Optional custom system prompt
            (default: SQL_GENERATOR_SYSTEM_PROMPT).
        :param user_prompt_template: Optional custom user prompt template
            (default: SQL_GENERATOR_USER_PROMPT_TEMPLATE).
        """
        self.llm_client = llm_client
        self.system_prompt = system_prompt or SQL_GENERATOR_SYSTEM_PROMPT
        self.user_prompt_template = (
            user_prompt_template or SQL_GENERATOR_USER_PROMPT_TEMPLATE
        )

    async def generate_code(
        self,
        query: str,
        intent: QueryIntent,
        context: RetrievalResult
    ) -> SQLQueryResult:
        user_prompt = self.user_prompt_template.format(
            query=query,
            schemas=context.schemas,
            examples=context.examples,
            documentation=context.documentation,
        )
        response = await self.llm_client.generate(
            self.system_prompt, user_prompt
        )

        try:
            response_json = json.loads(response)
            return SQLQueryResult(**response_json)
        except json.JSONDecodeError:
            return SQLQueryResult(sql_query=response)



from gru.agents.tools.core.code_analyzer.base import CodeAnalyzer
import ast
from typing import List


class ArgumentAnalyzer(CodeAnalyzer):
    """Python code analyzer that extracts command-line arguments."""
    
    class _ArgumentVisitor(ast.NodeVisitor):
        """AST visitor for extracting command-line arguments from Python code."""
        
        def __init__(self, tree, results_list):
            """
            Initialize visitor with the AST tree and a results list.
            
            Args:
                tree: The AST tree to analyze
                results_list: A list to which the visitor will append results
            """
            self.tree = tree
            self.results_list = results_list
        
        def is_argparse_constructor(self, node):
            """Check if a node is an argparse.ArgumentParser constructor."""
            return (
                isinstance(node.func, ast.Attribute)
                and isinstance(node.func.value, ast.Name)
                and node.func.value.id == "argparse"
                and node.func.attr == "ArgumentParser"
            )

        def extract_argument_values(self, node):
            """Extract argument values from a Call node's arguments."""
            return [arg.value for arg in node.args if isinstance(arg, ast.Constant)]

        def collect_add_arguments(self, tree):
            """Collect all add_argument calls in the AST tree."""
            for parent in ast.walk(tree):
                if (
                    isinstance(parent, ast.Call)
                    and isinstance(parent.func, ast.Attribute)
                    and parent.func.attr == "add_argument"
                ):
                    args = self.extract_argument_values(parent)
                    if args:
                        self.results_list.extend(args)

        def visit_Call(self, node):
            """Visit a Call node and check if it's an argparse constructor."""
            if self.is_argparse_constructor(node):
                self.collect_add_arguments(self.tree)
    
    def analyze(self, source_code: str) -> List[str]:
        """
        Analyze Python source code to extract command-line arguments.
        
        Args:
            source_code: Python source code as a string
            
        Returns:
            List of command-line arguments found in the code
        """
        tree = ast.parse(source_code)
        argparse_details = []
        
        visitor = self._ArgumentVisitor(tree, argparse_details)
        visitor.visit(tree)
        
        return argparse_details



from abc import ABC, abstractmethod
from typing import Any

class CodeAnalyzer(ABC):
    """Abstract base class for code analysis."""
    
    @abstractmethod
    def analyze(self, source_code: str) -> Any:
        """
        Analyze the provided source code.
        
        Args:
            source_code: Source code as a string
            
        Returns:
            Analysis results, type depends on the specific analyzer
        """
        pass



from typing import Dict
from urllib.parse import urljoin
import requests
from gru.sinks.sink import Sink
from gru.utils.config_reader import ConfigReader
from gru.utils.cloud_providers import CloudProviders
from gru.sinks.sink_family import SinkFamily



class OnlineFeatureStoreReader:
    def __init__(self, key, features, online_fs_servie_url, config_path: str = "./gru/config.yaml", online_feature_store: Sink = None):
        """
        A class representing a feature retrieval service with an online feature store.

        Args:
            key (str): A unique identifier for the feature retrieval request.
            features (list): A list of features to retrieve.
            config_path (str): The path to the configuration file used to initialize the reader.

        Returns:
            OnlineFeatureStoreReader: An instance of the OnlineFeatureStoreReader.

        Example:
            Initialize an OnlineFeatureStoreReader:
            reader = OnlineFeatureStoreReader(key="unique_key", features=["feature1", "feature2"], config_path="/path/to/config.yaml")
        """
        self.key = key
        self.features = features
        self.config = ConfigReader(config_path=config_path)
        if online_feature_store is None:       
            self.validate_config()
            self.online_feature_store = self.config.online_feature_store
        else:
            self.online_feature_store=online_feature_store.to_json()
        self.online_fs_service_url = online_fs_servie_url

    def validate_config(self):
        """
        Validates the family and cloud provider provider by the user in config.yaml
        """
        online_feature_store = self.config.online_feature_store
        try:
            SinkFamily.get_key(online_feature_store["family"])
            CloudProviders.get_key(online_feature_store["cloud_provider"])
        except Exception as e:
            raise e

    def to_json(self) -> Dict: 
        """
        Serialize the OnlineFeatureStoreReader instance to a JSON-compatible dictionary.
        """
        return {
            "key": self.key,
            "features": self.features,
            "online_fs_details": self.online_feature_store,
            "online_fs_service_url": self.online_fs_service_url,
        }

    def get_online_features(self):
        """
        Retrieve feature values for a particular key from an online feature store.

        Sends a GET request to the online feature store service endpoint specified in the configuration
        and retrieves the requested features for the provided key.

        Example:
            Retrieve online features for the initialized OnlineFeatureStoreReader instance:
            reader = OnlineFeatureStoreReader(key="unique_key", features=["feature1", "feature2"], config_path="/path/to/config.yaml")
            reader.get_online_features()
        """
        api_url = urljoin(
                        self.config.stuart_api_endpoint,
                        self.config.online_fs_service_url
        )
        result = requests.get(
            url=api_url,
            json=self.to_json(),
            headers={"Content-type": "application/json", "Accept": "text/plain"},
        )
        print(result.json())



from abc import ABC, abstractmethod
from typing import Any, Dict

from gru.agents.tools.core.vector_db.base import (
    AddToCollectionResponse,
    DeleteFromCollectionResponse,
    UpdateCollectionResponse,
    SimilaritySearchResponse
)


class BaseMemory(ABC):
    """Base class for memory implementations"""

    @abstractmethod
    async def store(self, data: Dict[str, Any], collection_name: str) -> AddToCollectionResponse:
        """Store data in specified collection"""
        pass

    @abstractmethod
    async def retrieve(
        self, 
        query: str, 
        collection_name: str,
        top_k: int = 5
    ) -> SimilaritySearchResponse:
        """Retrieve data from specified collection based on query"""
        pass

    @abstractmethod
    async def update(
        self, 
        doc_id: str, 
        data: Dict[str, Any],
        collection_name: str
    ) -> UpdateCollectionResponse:
        """Update existing data in specified collection"""
        pass

    @abstractmethod
    async def delete(
        self, 
        doc_id: str,
        collection_name: str
    ) -> DeleteFromCollectionResponse:
        """Delete data from specified collection"""
        pass


from abc import ABC, abstractmethod
from typing import Any


class Operation(ABC):
    """
    Abstract class defining the contract for all operations that can be run in different processing engines.

    This abstract class serves as a contract for defining operations that can be executed on various processing engines,
    providing a consistent interface for different implementations.

    Attributes:
    - kwargs (dict): Additional keyword arguments specific to the operation.


    Usage:
    ```
    # Example of creating a custom operation
    class CustomOperation(Operation):
        def __init__(self, custom_parameter, **kwargs):
            super().__init__(**kwargs)
            self.custom_parameter = custom_parameter

        def apply(self, df):
            # Custom operation implementation using  self.kwargs, and self.custom_parameter
            result = df  # Replace this with the actual implementation
            return result

    # Example usage
    custom_operation = CustomOperation(spark, custom_parameter_value, additional_parameter=...)
    result_df = custom_operation.apply(input_df)
    ```
    """

    def __init__(self, **kwargs):
        """
        Initialize the Operation with parameters specific to the operation.

        Parameters:
        - kwargs (dict): Additional keyword arguments specific to the operation.
        """
        self.kwargs = kwargs

    @abstractmethod
    def apply(self, *df: Any) -> Any:
        """
        Abstract method to apply the operation on a given data frame.

         Parameters:
        - *dfs (Any): Variable number of input data frames on which the operation will be applied.

        Returns:
        - Any: The result of applying the operation on the input data frame(s).
        """
        pass



from typing import Any
from pydantic import BaseModel


class AgentInvokeRequest(BaseModel):
    prompt_id: str
    prompt_body: dict[str, Any]


class AgentInvokeResponse(BaseModel):
    prompt_id: str


class TaskCompleteRequest(BaseModel):
    prompt_id: str
    tool_call_id: str
    task_type: str
    status: str
    result: dict[str, Any]


class AgentConversationRequest(BaseModel):
    conversation_id: str
    message: str


class AgentConversationResponse(BaseModel):
    message: str



"""
Create access token and update access token for client authentication.
"""
from typing import Dict


class GenerateAccessToken:
    def __init__(
        self,
        client_id,
        user_email,
        password,
    ):
        self.client_id = client_id
        self.user_email = user_email
        self.password = password

    def to_json(self) -> Dict:
        """
        Represent generate token Object
        Parameters:
        ----------
        client_id (str):
            client id for user.
        user_email (str):
            user email of user.
        password (str):
            password of user.
        """
        return {
            "client_id": self.client_id,
            "user_email": self.user_email,
            "password": self.password,
        }



from datetime import datetime
from typing import List, Optional, Union

from pydantic import BaseModel
from gru.schema.api_response_handler import ApiError
from gru.risk.workflows.apis.workflows import create_workflow, deploy_workflow, get_workflow, get_workflows, update_workflow_status
from gru.risk.workflows.models.workflows import (
    CreateWorkflowWithRuleRequest, 
    RuleDetails, 
    UpdateWorkflowStatusResponse, 
    Workflow, 
    WorkflowDeployment, 
    WorkflowDetails
)
from gru.risk.workflows.rule_commands import RulesCommands
from gru.risk.workflows.utils import read_json_config
from gru.cli import read_token
from gru.utils.config_reader import ConfigReader

def format_timestamp(timestamp: datetime) -> str:
    """Format timestamp to a readable string"""
    return timestamp.strftime("%Y-%m-%d %H:%M:%S UTC")

class WorkflowCommands(object):

    def __init__(self):
        self.rules = RulesCommands()

    def _format_workflow_deployment(self, deploy: WorkflowDeployment) -> str:
        """Format a single workflow deployment"""
        cluster = deploy.cluster_name
        status = deploy.deployment_status
        last_synced = format_timestamp(deploy.last_synced_at) if deploy.last_synced_at else 'Never'
        return f" - {cluster}: {status} (Last synced: {last_synced})"

    def _format_workflow(self, workflow: Union[Workflow, UpdateWorkflowStatusResponse]) -> List[str]:
        """Format a workflow into a list of output lines"""
        output_lines = []
        
        status_str = f"[{workflow.status}]" if workflow.status else ""
        desc_str = f": {workflow.description}" if workflow.description else ""

        deploy_str = ""

        if hasattr(workflow, 'deployments') and workflow.deployments:
            deploys = len(workflow.deployments)
            deploy_str = f" ({deploys} deployment{'s' if deploys != 1 else ''})"
        
        output_lines.append(
            f"{workflow.workflow_name} {status_str}{desc_str}{deploy_str}"
        )

        if hasattr(workflow, 'deployments') and workflow.deployments:
            output_lines.append(" Deployments:")
            for deploy in workflow.deployments:
                output_lines.append(self._format_workflow_deployment(deploy))
            output_lines.append("")  
            
        return output_lines

    def _format_rule(self, rule: RuleDetails) -> List[str]:
        """Format a single rule into a list of output lines"""
        output_lines = []
        output_lines.append(f" - {rule.rule_name} [{rule.status}] (Stage: {rule.stage})")
        return output_lines

    def _format_workflow_detailed(self, workflow: WorkflowDetails) -> List[str]:
        """Format a workflow with rules into a list of output lines"""
        output_lines = self._format_workflow(workflow)
        
        if workflow.rules:
            output_lines.append(" Rules:")
            for rule in workflow.rules:
                output_lines.extend(self._format_rule(rule))
        
        return output_lines
    
    def _format_model_response(self, response: BaseModel) -> str:
        """Format a model response into a string"""
        return "\n".join(f"  {k}: {v}" for k, v in response.model_dump().items())

    def list(
    self, 
    workflow_name: Optional[str] = None, 
    is_active: Optional[bool] = None,
    detailed: bool = False
    ) -> str:
        """
        List workflows or get details of a specific workflow.
        
        Args:
            workflow_name: Optional name of specific workflow to retrieve
            is_active: Optional filter for active/inactive workflows
            detailed: Whether to show detailed information including rules
            
        Returns:
            Formatted string containing workflow information
        """
        try:
            auth_token = read_token()
            
            if workflow_name:
                workflow = get_workflow(auth_token, workflow_name)
                return "\n".join(self._format_workflow_detailed(workflow))
            

            workflows = get_workflows(auth_token, is_active)
            if not workflows:
                return "No workflows found"
            
            output_lines = []
            for workflow in workflows:
                if detailed:
                    detailed_workflow = get_workflow(auth_token, workflow.workflow_name)
                    output_lines.extend(self._format_workflow_detailed(detailed_workflow))
                else:
                    output_lines.extend(self._format_workflow(workflow))
                
            return "\n".join(output_lines).rstrip()
            
        except ApiError as api_error:
            return f"API Error: {api_error.message}"
        except ValueError as value_error:
            return str(value_error)
        except Exception as e:
            return f"Error: {str(e)}"
    
    def create(self, workflow_name: str, config: str) -> str:
        """
        Create a new workflow using configuration from JSON file.
        
        Args:
            workflow_name: Name for the new workflow
            config: Path to JSON configuration file
            
        Returns:
            Formatted string showing created workflow details
        """
        try:
            auth_token = read_token()
            
            config_data = read_json_config(config)
            
            config_data["workflow_name"] = workflow_name
            
            request = CreateWorkflowWithRuleRequest.model_validate(config_data)
            
            workflow = create_workflow(auth_token, request)
            
            return "\n".join([
                f"Successfully created workflow: {workflow_name}",
                "",
                *self._format_workflow_detailed(workflow)
            ])
            
        except ApiError as api_error:
            return f"API Error: {api_error.message}"
        except ValueError as value_error:
            return f"Configuration error: {str(value_error)}"
        except Exception as e:
            return f"Error: {str(e)}"
    
    def update_status(self, workflow_name: str, status: str) -> str:
        """
        Update the status of a workflow.
        
        Args:
            workflow_name: Name of the workflow to update
            status: New status ('ACTIVE', 'INACTIVE', or 'DEPRECATED')
            
        Returns:
            Formatted string showing updated workflow details
        """
        try:
            auth_token = read_token()
            
            workflow = update_workflow_status(auth_token, workflow_name, status)
            
            return "\n".join([
                f"Successfully updated workflow status:",
                "",
                *self._format_workflow(workflow)
            ])
            
        except ApiError as api_error:
            return f"API Error: {api_error.message}"
        except ValueError as value_error:
            return f"Invalid status: {str(value_error)}"
        except Exception as e:
            return f"Error: {str(e)}"
        
    def deploy(
    self,
    workflow_name: str,
    env: str,
    cluster_name: str,
    namespace: str,
    env_vars: str = "",
    hpa_configs: Optional[str] = None
    ) -> str:
        """
        Deploy a workflow to the specified environment.
        
        Args:
            workflow_name: Name of the workflow to deploy
            env: Target environment (STAGING/PRODEXPERIMENT/LIVE)
            cluster_name: Name of the target cluster
            namespace: Kubernetes namespace
            env_vars: Optional environment variables in format "key1=value1,key2=value2"
            hpa_configs: Optional path to HPA configuration file (YAML/JSON)
            
        Returns:
            Formatted string showing deployment result
        """
        try:
            auth_token = read_token()
            
            env_vars_dict = {}
            if env_vars:
                try:
                    env_vars_dict = dict(
                        pair.split('=', 1) 
                        for pair in env_vars.split(',')
                    )
                except ValueError:
                    return "Error: env_vars must be in format 'key1=value1,key2=value2'"
            
            hpa_config = ConfigReader.load_hpa_config(hpa_configs)

            env = env.upper()
            if env not in ["BACKTESTING", "STAGING", "PRODEXPERIMENT", "LIVE"]:
                return "Error: env must be one of: BACKTESTING, STAGING, PRODEXPERIMENT, LIVE"
            
            response = deploy_workflow(
                auth_token,
                workflow_name,
                cluster_name,
                env,
                namespace,
                env_vars_dict,
                hpa_config
            )
            
            return "\n".join([
                f"Successfully queued workflow deployment:",
                "",
                f"Details: ",
                self._format_model_response(response)
            ])
            
        except ApiError as api_error:
            return f"API Error: {api_error.message}"
        except ValueError as value_error:
            return f"Invalid input: {str(value_error)}"
        except Exception as e:
            return f"Error: {str(e)}"





import yaml
from typing import Dict, Any, List, Optional

from gru.agents.tools.core.vector_db.base import VectorDBClient

class CollectionManager:
    """
    Manages vector database collections, handling initialization and schema management.
    This class extracts collection handling logic from CansoMemory for better separation
    of concerns and readability.
    """
    
    def __init__(self, vector_db_client: VectorDBClient, embedding_dimension: int, config_path: str):
        """
        Initialize the collection manager.
        
        Args:
            vector_db_client: The vector database client to use
            embedding_dimension: Dimension for embedding vectors
            config_path: Path to collections configuration file
        """
        self.vdb_client = vector_db_client
        self.embedding_dimension = embedding_dimension
        self.collections_config = self._load_config(config_path)
    
    #TODO: Use pydantic to validate the config
    def _load_config(self, config_path: str) -> Optional[Dict[str, Any]]:
        """Load collections configuration from YAML file"""
        with open(config_path, 'r') as file:
            config = yaml.safe_load(file)
            return config
    
    def initialize_collections(self) -> None:
        """Initialize collections defined in the configuration file"""
        if not self.collections_config:
            print("No collections configuration found")
            return

        collections = self.collections_config.get("collections", {})
        print(f"Initializing {len(collections)} collections")
        
        for collection_name, collection_config in collections.items():
            try:
                print(f"Creating collection: {collection_name}")
                schema = self._build_collection_schema(collection_config)
                
                replace_existing = collection_config.get("replace_existing", False)
                
                self.vdb_client.create_collection_sync(
                    collection_name=collection_name,
                    schema=schema,
                    replace_existing=replace_existing
                )
            except Exception as e:
                print(f"Warning: Failed to create collection {collection_name}: {str(e)}")
    
    def _build_collection_schema(self, collection_config: Dict[str, Any]) -> Dict[str, Any]:
        """Build schema for a collection based on configuration"""
        fields = self._prepare_fields(collection_config)
        
        schema = {
            "fields": fields,
            "description": collection_config.get("description", ""),
            "create_indexes": collection_config.get("create_indexes", [])
        }
        
        return schema
    
    def _prepare_fields(self, collection_config: Dict[str, Any]) -> List[Dict[str, Any]]:
        """Process and prepare collection fields"""
        fields = []
        
        for field in collection_config.get("fields", []):
            field_copy = field.copy()
            
            if field_copy.get("name") == "embedding" and field_copy.get("dtype") == "FLOAT_VECTOR":
                field_copy["dim"] = self.embedding_dimension
                
            fields.append(field_copy)
        
        fields = self._ensure_required_fields(fields)
        
        return fields
    
    def _ensure_required_fields(self, fields: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Ensure required fields exist in the schema"""
        if not any(field.get("name") == "id" for field in fields):
            fields.append({
                "name": "id",
                "dtype": "VARCHAR",
                "is_primary": True,
                "max_length": 100
            })
        
        if not any(field.get("name") == "embedding" and field.get("dtype") == "FLOAT_VECTOR" 
                  for field in fields):
            fields.append({
                "name": "embedding",
                "dtype": "FLOAT_VECTOR",
                "dim": self.embedding_dimension
            })
        
        return fields
        
    async def get_collection_info(self, collection_name: str) -> Dict[str, Any]:
        """Get information about a collection"""
        try:
            collections = await self.vdb_client.list_collections()
            if collection_name not in collections:
                return {"status": "error", "message": "Collection not found"}
            
            return await self.vdb_client.get_collection_info(collection_name)
        except Exception as e:
            print(f"Error fetching collection info: {str(e)}")
            return {"status": "error", "message": str(e)}


"""
[Not yet Implemented] Sink Module for a MySQL Database
"""

# from gru.sinks import sink
# from typing import Dict

# class MySQLDBSink(sink):
#     """A MySQL Database sink"""

#     def __init__(
#         self,
#         host : str,
#         username : str,
#         password : str,
#         database : str,
#         table: str,
#         port : int,
#         write_properties : Dict
#     ) -> None:
#         super().__init__(
#             name = "",
#             family = "mysql_db",
#             physical_uri = {
#                 "host": host,
#                 "username": username,
#                 "password": password,
#                 "database": database,
#                 "port": port,
#             }
#         )

#         self.host = host
#         self.username = username
#         self.password = password
#         self.database = database
#         self.table = table
#         self.port = port
#         self.write_properties = {} if write_properties is None else write_properties
        

#     def to_json(self) -> Dict:
#         """
#         :return: The JSON representation of the object.
#         """
#         return {
#             "host" : self.host,
#             "username" : self.username,
#             "password" : self.password,
#             "database" : self.database,
#             "table" : self.table,
#             "port" : self.port
#         }
    
#     @classmethod
#     def from_json(cls, json_dict: Dict) -> "MySQLDBSink":
#         """
#         Create a MySQLDBSink instance from a JSON dictionary.

#         Args:
#             json_dict (Dict): A dictionary containing sink parameters

#         Returns:
#             MySQLDBSink: An instance of MySQLDBSink
#         """
#         return MySQLDBSink(
#             host=json_dict["host"],
#             username=json_dict["username"],
#             password=json_dict["password"],
#             database=json_dict["database"],
#             port=json_dict["port"],
#             table=json_dict["table"],
#         )



import os
import uuid
import fire
import subprocess
import json

from gru.agents.apis import (
    AGENT_CONFIG_FILE_NAME,
    ai_agent_templates_setup,
    converse_agent,
    delete_agent,
    deploy_agent,
    read_prompt_file,
    register_agent,
    update_agent,
    memory_insert,
    memory_update,
    memory_delete,
)
from gru.agents.apis import prompt_agent
from gru.components.apis import setup as component_setup
from gru.schema.api_response_handler import ApiError
from gru.utils.constants import TOKEN_ENV_VAR_NAME


def get_env_token() -> str:
    auth_token = os.getenv(TOKEN_ENV_VAR_NAME)
    if auth_token == None:
        raise ValueError(f"Environment variable {TOKEN_ENV_VAR_NAME} missing")
    return auth_token


def create_correlation_id() -> str:
    return str(uuid.uuid4())


class ComponentCommands(object):
    def __init__(self, auth_token=None):
        self._auth_token = auth_token

    def get_token(self):
        return self._auth_token if self._auth_token is not None else get_env_token()

    def setup(self, cluster_name: str, config_file: str):
        correlation_id = create_correlation_id()
        try:
            auth_token = self.get_token()
            result = component_setup(
                correlation_id, cluster_name, config_file, auth_token
            )
            return str(result)
        except FileNotFoundError:
            return f"Error: {config_file} file not found."
        except ValueError as value_error:
            return str(value_error)
        except Exception:
            return f"An unexpected error occured. Correlation ID: {correlation_id}"


class AgentCommands(object):
    def __init__(self, auth_token=None):
        self._auth_token = auth_token
        self.memory = MemoryCommands(auth_token)

    def get_token(self):
        return self._auth_token if self._auth_token is not None else get_env_token()

    def create_bootstrap(self):
        try:
            ai_agent_templates_setup()
            return f"Agent bootstrap project created successfully!"
        except subprocess.CalledProcessError as value_error:
            return str(value_error)

    def register(self, agent_folder, cluster_name, image, image_pull_secret):
        correlation_id = create_correlation_id()
        try:
            auth_token = self.get_token()
            result = register_agent(
                correlation_id,
                auth_token,
                agent_folder,
                cluster_name,
                image,
                image_pull_secret,
            )
            return str(result)
        except FileNotFoundError:
            return f"Error: {os.path.join(agent_folder, AGENT_CONFIG_FILE_NAME)} file not found."
        except ValueError as value_error:
            return str(value_error)
        except Exception:
            return f"An unexpected error occured. Correlation ID: {correlation_id}"

    def deploy(self, agent_name):
        correlation_id = create_correlation_id()
        try:
            auth_token = self.get_token()
            result = deploy_agent(correlation_id, auth_token, agent_name)
            return str(result)
        except ValueError as value_error:
            return str(value_error)
        except Exception:
            return f"An unexpected error occured. Correlation ID: {correlation_id}"

    def update(self, agent_folder, image=None, image_pull_secret=None):
        correlation_id = create_correlation_id()
        try:
            auth_token = self.get_token()
            return update_agent(
                correlation_id, auth_token, agent_folder, image, image_pull_secret
            )
        except FileNotFoundError:
            return f"Error: {os.path.join(agent_folder, AGENT_CONFIG_FILE_NAME)} file not found."
        except ValueError as value_error:
            return str(value_error)
        except Exception:
            return f"An unexpected error occured. Correlation ID: {correlation_id}"

    def delete(self, agent_name: str):
        correlation_id = create_correlation_id()
        try:
            auth_token = self.get_token()
            return delete_agent(correlation_id, auth_token, agent_name)
        except ValueError as value_error:
            return str(value_error)
        except Exception:
            return f"An unexpected error occured. Correlation ID: {correlation_id}"

    def prompt(self, agent_name: str, prompt_file: str) -> str:
        """
        Send a prompt to a deployed agent.

        Args:
            agent_name (str): Name of the deployed agent
            prompt_file (str): Path to JSON file containing the prompt

        Returns:
            str: Success or error message
        """
        try:
            auth_token = self.get_token()
            prompt_data = read_prompt_file(prompt_file)
            result = prompt_agent(agent_name, prompt_data, auth_token)
            return str(result)
        except FileNotFoundError as e:
            return f"File error: {str(e)}"
        except json.JSONDecodeError as e:
            return f"JSON error: {str(e)}"
        except ApiError as e:
            return f"API error: {e.message}"
        except ValueError as e:
            return f"Value error: {str(e)}"
        except Exception as e:
            return f"Unexpected error: {str(e)}"

    def converse(self, agent_name: str, conversation_id: str | None = None):
        try:
            auth_token = self.get_token()
            return converse_agent(auth_token, agent_name, conversation_id)
        except ValueError as value_error:
            return str(value_error)
        except Exception:
            return "An unexpected error occured."


class MemoryCommands(object):
    def __init__(self, auth_token=None):
        self._auth_token = auth_token

    def get_token(self):
        return self._auth_token if self._auth_token is not None else get_env_token()

    def insert(self, agent_name: str, file: str):
        """
        Insert data into agent memory
        Args:
            agent_name: Name of the agent
            file: JSON file containing memory data
        Returns:
            Success or error message
        """
        correlation_id = create_correlation_id()
        try:
            auth_token = self.get_token()
            result = memory_insert(correlation_id, auth_token, agent_name, file)
            return str(result)
        except FileNotFoundError:
            return f"Error: {file} file not found."
        except json.JSONDecodeError as e:
            return f"JSON error: {str(e)}"
        except ValueError as value_error:
            return str(value_error)
        except Exception as e:
            return f"An unexpected error occurred. Correlation ID: {correlation_id}. Error: {str(e)}"

    def update(self, agent_name: str, file: str):
        """
        Update data in agent memory
        Args:
            agent_name: Name of the agent
            file: JSON file containing update data
        Returns:
            Success or error message
        """
        correlation_id = create_correlation_id()
        try:
            auth_token = self.get_token()
            result = memory_update(correlation_id, auth_token, agent_name, file)
            return str(result)
        except FileNotFoundError:
            return f"Error: {file} file not found."
        except json.JSONDecodeError as e:
            return f"JSON error: {str(e)}"
        except ValueError as value_error:
            return str(value_error)
        except Exception as e:
            return f"An unexpected error occurred. Correlation ID: {correlation_id}. Error: {str(e)}"

    def delete(self, agent_name: str, memory_id: str, collection: str):
        """
        Delete data from agent memory
        Args:
            agent_name: Name of the agent
            memory_id: ID of the memory to delete
            collection: Collection name containing the memory
        Returns:
            Success or error message
        """
        correlation_id = create_correlation_id()
        try:
            auth_token = self.get_token()
            result = memory_delete(
                correlation_id, auth_token, agent_name, memory_id, collection
            )
            return str(result)
        except ValueError as value_error:
            return str(value_error)
        except Exception as e:
            return f"An unexpected error occurred. Correlation ID: {correlation_id}. Error: {str(e)}"


class GruCommands(object):
    def __init__(self):
        self.component = ComponentCommands()
        self.agent = AgentCommands()


def main():
    fire.Fire(GruCommands)



from abc import ABC, abstractmethod
from typing import Any

from gru.agents.schemas import AgentInvokeRequest
from gru.agents.schemas.schemas import (
    AgentConversationRequest,
    AgentConversationResponse,
    TaskCompleteRequest,
)


class AgentWorkflow(ABC):

    @abstractmethod
    async def setup(self):
        pass

    @abstractmethod
    async def invoke(self, request: AgentInvokeRequest) -> dict[str, Any]:
        pass

    @abstractmethod
    async def resume(self, request: TaskCompleteRequest) -> dict[str, Any]:
        pass

    @abstractmethod
    async def converse(
        self, request: AgentConversationRequest
    ) -> AgentConversationResponse:
        pass



# from typing import List, Any
# from sentence_transformers import SentenceTransformer
# from gru.agents.tools.core.embeddings.base import EmbeddingGenerator
#
#
# class SentenceTransformerEmbeddingGenerator(EmbeddingGenerator):
#     MODELS = {
#         "all-MiniLM-L6-v2": 384,
#         "all-mpnet-base-v2": 768,
#         "paraphrase-multilingual-MiniLM-L12-v2": 384
#     }
#
#     def __init__(
#             self,
#             client: Any,
#             model: str = "all-MiniLM-L6-v2",
#             batch_size: int = 32
#     ):
#         self.client = client
#         if model not in self.MODELS:
#             raise ValueError(f"Model {model} not supported. Choose from {list(self.MODELS.keys())}")
#
#         self.model_name = model
#         self.model = SentenceTransformer(model)
#         self.batch_size = batch_size
#         self._dimension = self.MODELS[model]
#
#     @property
#     def dimension(self) -> int:
#         return self._dimension
#
#     async def generate(self, text: str) -> List[float]:
#         """Generate embeddings for a single text string"""
#         embedding = self.model.encode(text)
#         return embedding.tolist()
#
#     async def generate_batch(self, texts: List[str]) -> List[List[float]]:
#         """Generate embeddings for a batch of text strings"""
#         all_embeddings = []
#         for i in range(0, len(texts), self.batch_size):
#             batch = texts[i:i + self.batch_size]
#             embeddings = self.model.encode(batch)
#             # Convert numpy arrays to lists for serialization
#             embeddings_list = [emb.tolist() for emb in embeddings]
#             all_embeddings.extend(embeddings_list)
#         return all_embeddings


import json
from pathlib import Path
from typing import Dict, Any

def read_json_config(file_path: str) -> Dict[str, Any]:
    """
    Read and validate JSON configuration file.
    
    Args:
        file_path: Path to JSON config file
        
    Returns:
        Dict containing parsed JSON data
        
    Raises:
        ValueError: If file doesn't exist or isn't valid JSON
    """
    path = Path(file_path)
    if not path.exists():
        raise ValueError(f"Config file not found: {file_path}")
        
    try:
        with path.open() as f:
            return json.load(f)
    except json.JSONDecodeError as e:
        raise ValueError(f"Invalid JSON in config file: {str(e)}")


from enum import Enum


class ProcessingEngine(Enum):
    """
    Enum class representing processing engines.
    """

    PYSPARK_K8S = "pyspark_k8s"
    PYTHON = "python"
    DASK = "dask"

    @classmethod
    def get_key(cls, value: str):
        if value == cls.PYSPARK_K8S.value:
            return cls.PYSPARK_K8S
        elif value == cls.PYTHON.value or value == cls.DASK.value:
            raise ValueError(
                f"{value} isn't supported yet. Support expected to be available around Feb, 2024."
            )
        else:
            raise ValueError(
                f"""Unexpected engine {value}. Supported Engines are {', '.join(engine.value for engine in ProcessingEngine)}."""
            )



import os
from typing import Optional, Type
from pydantic import BaseModel, Field
from langchain_core.runnables import RunnableConfig
from langchain_core.callbacks import CallbackManagerForToolRun
from langchain_core.tools import BaseTool

from gru.agents.checkpoint.task_results import TaskResultsRepository, TaskStatus


class PythonRunStatusToolInput(BaseModel):
    run_id: str = Field(description="run_id of the python code run")


class PythonRunStatusChecker(BaseTool):
    name: str = "get_python_run_status"
    description: str = "use this to get the status of a python code run"
    args_schema: Type[BaseModel] = PythonRunStatusToolInput
    return_direct: bool = True

    class Config:
        extra = "allow"

    def __init__(self):
        task_results_repo = TaskResultsRepository()
        super().__init__(task_results_repo=task_results_repo)
        self.task_results_repo = task_results_repo

    def _run(self, *args, **kwargs):
        return super()._run(*args, **kwargs)

    async def _arun(
        self,
        run_id: str,
        config: RunnableConfig,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ):
        agent_name = os.getenv("AGENT_NAME")
        result = await self.task_results_repo.get_result(
            agent_name, config["metadata"]["thread_id"], "run_python_code", run_id
        )

        if result:
            status, result_text = result
            if status == TaskStatus.PROCESSING.value:
                return status
            elif status == TaskStatus.COMPLETED.value:
                return result_text
        else:
            return "Python run task result not found"



from pydantic import BaseModel

class Broker(BaseModel):
    type: str
    name: str

class CheckpointDB(BaseModel):
    type: str
    name: str
    size: str

class VectorDB(BaseModel):
    type: str
    name: str
    size: str
    image_pull_secret: str

class TaskServer(BaseModel):
    type: str
    name: str
    replicas: int
    concurrency_per_replica: int
    broker_name: str
    aws_role_arn: str | None = None

class ResourceSetupRequest(BaseModel):
    cluster_name: str = None
    broker: Broker = None
    checkpoint_db: CheckpointDB = None
    task_server: TaskServer = None
    vector_db: VectorDB = None


from pydantic import BaseModel
from typing import Dict, Any, Optional


class AgentPromptRequest(BaseModel):
    """
    Model representing a prompt request to an agent.

    Attributes:
        prompt (Dict[str, Any]): Dictionary containing the prompt data
    """

    prompt: Dict[str, Any]


class AgentRegisterRequest(BaseModel):
    cluster_name: str
    agent_name: str
    image: str
    image_pull_secret: str
    task_server_name: str
    checkpoint_db_name: str
    replicas: int
    iam_role_arn: Optional[str] = None
    vector_db_name: Optional[str] = None


class AgentUpdateRequest(BaseModel):
    image: Optional[str] = None
    image_pull_secret: Optional[str] = None
    task_server_name: Optional[str] = None
    checkpoint_db_name: Optional[str] = None
    replicas: Optional[int] = None
    iam_role_arn: Optional[str] = None
    vector_db_name: Optional[str] = None


class MemoryInsertRequest(BaseModel):
    """
    Model for inserting data into memory.
    Attributes:
        collection_name: Name of the collection to insert data into
        text: Text content to be embedded
        metadata: Additional metadata for the memory
    """

    collection_name: str
    text: str
    data: Dict[str, Any]


class MemoryUpdateRequest(BaseModel):
    """
    Model for updating data in memory.
    Attributes:
        collection_name: Name of the collection containing the data
        memory_id: ID of the document to update
        text: New text content to be embedded
        metadata: Updated metadata for the memory
    """

    collection_name: str
    memory_id: str
    text: str
    data: Dict[str, Any]


class MemoryDeleteRequest(BaseModel):
    """
    Model for deleting data from memory.
    Attributes:
        collection_name: Name of the collection containing the data
        memory_id: ID of the document to delete
    """

    collection_name: str
    memory_id: str



"""
Utils Module consists of all helper functions and classes
"""


def check_data_type(field_datatypes_dict: dict) -> bool:
    """
    To ensure that the datatype of the value passed by the end user is correct

    Parameters
    ----------
    field_datatypes_dict:
        dict with datatypes as keys and value is the list of fields accepted as that datatype

    Returns
    -------
    bool:
        True, if encountered a field with incorrect datatype
        False, if all fields are of correct datatype

    Notes
    -----
    Even though the function returns None, still it prints success/failure message

    Examples
    --------
    """
    args_type_check = True
    for data_type, fields in field_datatypes_dict.items():
        for field in fields:
            if not isinstance(field, data_type):
                print(str(field), f"is supposed to be {data_type}")
                args_type_check = False
                return args_type_check

    return args_type_check


class Constants:
    """
    Constants Class to declare all the constant values that will be used across all modules
    """

    # Stuart API
    STUART_API_ENDPOINT = "http://35.154.143.223/"  # prod API endpoint

    # Infra
    API_DEPLOY_INFRA_DEV = STUART_API_ENDPOINT + "deploy/infrastructure"
    API_DEPLOY_INFRA_PROD = STUART_API_ENDPOINT + "deploy/infrastructure"

    # Infra
    REGISTER_INFRA_URL = STUART_API_ENDPOINT + "register/infrastructure"

    # Generate Training Data
    GTD_STAGE = "dev"
    GTD_JOB_TIMEOUT = 30
    GTD_REQUEST_ID = "01"
    GTD_REQUEST_TYPE = "gtd"
    TRAINING_DATA = "TRAINING_DATA"
    GTD_K8S_NAMESPACE = "airflow-jobs"
    GTD_REGISTER_URL = STUART_API_ENDPOINT + "training_data/register"
    GTD_GENERATE_DATA_URL = STUART_API_ENDPOINT + "training_data/generate"

    AIRFLOW_DAGS_BACKUP_PATH = "s3://internal-ml-demos/dags/"

    def __init__(self) -> None:
        pass


from enum import Enum

class DataProcessingParadigm(Enum):
    """
    Enum class representing various programming paradigms.
    """

    Batch = "batch"
    Streaming = "streaming"


    @classmethod
    def get_key(cls, value: str) -> 'DataProcessingParadigm':
        """
        Get the enum value for a given string representation.

        Args:
            paradigm_str (str): A string representing a programming paradigm.

        Returns:
            Paradigms: The corresponding enum value if found, or None if not found.
        """
        for paradigm in cls:
            if paradigm.value == value:
                return paradigm
        raise ValueError(f"No matching enum for value: {value}")


from typing import Dict, Union
from datetime import datetime
from gru.sinks.sink import Sink
from gru.redis_helpers.elasticache import ElasticCacheRedisClusterMode
from gru.sinks.sink_family import SinkFamily
from gru.utils.cloud_providers import CloudProviders


class ElasticCacheRedisSink(Sink):
    """Sink module for Redis on AWS ElasticCache."""

    def __init__(
        self,
        name: str,
        host: str,
        port: int = 6379,
        mode: Union[
            ElasticCacheRedisClusterMode, str
        ] = ElasticCacheRedisClusterMode.CLUSTER_ENABLED,
        created_at=datetime.now(),
        **kwargs,
    ):
        """
        Initialize the ElasticCacheRedisSink.

        Arguments:
        - host (str): Primary Endpoint of the ElasticCache Redis Cluster.
        - port (int): The port of the Redis instance to connect. Default is 6379.
        - mode (Union[ElasticCacheRedisClusterMode, str]): Whether Cluster mode is enabled or Cluster mode is disabled.
          Default is ElasticCacheRedisClusterMode.CLUSTER_ENABLED.

        Raises:
        - ValueError: If host is not a string or port is not an integer.
        """
        super().__init__(
            name=name,
            family=SinkFamily.REDIS.value,
            cloud_provider=CloudProviders.AWS,
            physical_uri={"host": host, "port": port, "mode": mode.value},
            type="ElasticCacheRedisSink",
            **kwargs,
        )
        self.host = host
        self.port = port
        self.mode = (
            mode
            if isinstance(mode, ElasticCacheRedisClusterMode)
            else ElasticCacheRedisClusterMode(mode)
        )

        self._validate_host_port()
        self._validate_mode()
        self.created_at = created_at

    def _validate_host_port(self):
        """
        Validate that host is a string and port is an integer.

        Raises:
        - ValueError: If host is not a string or port is not an integer.
        """
        if not isinstance(self.host, str):
            raise ValueError("Host must be a string.")
        if not isinstance(self.port, int):
            raise ValueError("Port must be an integer.")

    def _validate_mode(self):
        """
        Validate the mode value.

        Raises:
        - ValueError: If mode is not an instance of ElasticCacheRedisClusterMode or if mode is missing in JSON data.
        """
        if not isinstance(self.mode, ElasticCacheRedisClusterMode):
            raise ValueError(
                "Mode must be an instance of ElasticCacheRedisClusterMode."
            )
        if self.mode is None:
            raise ValueError("Missing 'mode' value in JSON data.")

    def to_json(self) -> Dict:
        """
        Convert the ElasticCacheRedisSink instance to a JSON-serializable dictionary.

        Returns:
        - Dict: A dictionary containing the name, family, host, port, and mode values.
        """
        return {
            "name": self.name,
            "family": self.family,
            "physical_uri": self.physical_uri,
            "type": self.type,
            "file_type": self.file_type.type if self.file_type else None,
            "file_type_properties": self.file_type.options if self.file_type else None,
            "cloud_provider": self.cloud_provider.value,
            "description": self.description,
            "metadata": self.metadata,
            "owners": self.owners,
            "created_at": str(self.created_at),
        }

    @classmethod
    def from_json(cls, json_dict: Dict) -> "ElasticCacheRedisSink":
        """
        Create an ElasticCacheRedisSink instance from a JSON dictionary.

        Arguments:
        - json_dict (Dict): A dictionary containing the host, port, and mode values.

        Returns:
        - ElasticCacheRedisSink: An instance of ElasticCacheRedisSink.

        Raises:
        - ValueError: If host is missing in the JSON data.
        """
        host = json_dict.get("host")
        port = json_dict.get("port")
        mode = json_dict.get("mode")

        return ElasticCacheRedisSink(host=host, port=port, mode=mode)



SQL_GENERATOR_SYSTEM_PROMPT = """
You are a highly intelligent assistant specialized in generating SQL queries. Your task is to analyze the user query and generate a correct, optimized SQL query in **PostgreSQL** syntax based on the provided database information. The output must be a JSON object with the following structure:

{
    "sql_query": "The generated SQL query",
    "tables_used": ["List of tables used in the query"],
    "columns_used": ["List of columns used in the query"],
    "join_conditions": ["List of join conditions if any"]
}

**Key Considerations:**
1. **Understand the User Query:**
   - Parse the query to identify the intent and required data.
2. **Analyze Database Information:**
   - Identify relevant tables, columns, and relationships (e.g., foreign keys, joins).
   - **Preserve the exact table names as provided.**
3. **Generate the Query:**
   - Use appropriate SQL commands (`SELECT`, `INSERT`, `UPDATE`, `DELETE`, etc.).
   - Include filters, joins, grouping, or sorting as needed.
   - Ensure the query adheres to the schema and data types.
4. **Validate Query Structure:**
   - Ensure the SQL query is syntactically correct and executable.
   - Optimize for performance by minimizing unnecessary joins and filters.

**Response Guidelines:**
1. **Context Sufficiency:**
   - If the provided context is sufficient, generate a valid SQL query without any explanations.
   - Ensure the query is fully executable and adheres to the provided schema and constraints.
2. **Context Insufficiency:**
   - If the provided context is insufficient, explicitly state what information is missing (e.g., missing tables, columns, or relationships).
   - Do not generate a query if critical information is missing.
3. **Query Accuracy:**
   - Double-check that the query accurately reflects the user's intent and retrieves the correct data.
   - Ensure the query uses the correct table relationships (e.g., joins) and filters.
4. **Syntax Compliance:**
   - Ensure the query strictly follows the specified syntax and best practices.
5. **Optimization:**
   - Optimize the query for performance by minimizing redundant operations and leveraging indexes where applicable.
6. **Error Handling:**
   - If the query cannot be generated due to ambiguity or insufficient information, provide a clear and concise explanation.

Return only the JSON object without any additional explanation.
"""

SQL_GENERATOR_USER_PROMPT_TEMPLATE = """
Generate a SQL query for the following request:

**Natural Language Query:**
{query}

**Available Tables and Schemas:**
{schemas}

**Documentation:**
{documentation}

**Relevant Examples:**
{examples}

Generate a single, optimized SQL query and return the output in the specified JSON format. Ensure the SQL query is executable and does not contain placeholders.
"""


class RequestHandler():
    """
    Class for handling requests.

    Methods:
        - parse_request: Abstract method for parsing requests.
    """
    def parse_request(self, request):
        """
        Parse incoming requests.

        Parameters:
            - request: Request object received from the web framework.

        Returns:
            dict: Parsed request data.
        """
        pass


"""
Deploy Infrastructure Components for different Data & ML Use Cases
This includes 
- A Managed Kubernetes Custer (EKS in AWS)
- Airflow
- Grafana
- Prometheus
- Loki
- Managed Postgres Database (RDS in AWS)
"""
import requests
from gru.utils.utils import Constants


class DeployInfrastructure:
    def __init__(
        self, deployment_id: str, access_token: str,
    ):
        self.deployment_id = deployment_id
        self.access_token = access_token

        self.deploy_infra_details = {
            "deployment_id": self.deployment_id,
            "access_token": self.access_token,
        }

    def deploy_infra(self):
        """deploy/spin-up infra."""
        result = requests.get(
            url=Constants.API_DEPLOY_INFRA_PROD,
            json=self.deploy_infra_details,
            headers={"Content-type": "application/json", "Accept": "text/plain"},
        )
        print(result.json())



from typing import Any, Dict, List, Literal, Optional
from gru.cli import read_token
from gru.risk.workflows.utils import read_json_config
from gru.schema.api_response_handler import ApiError
from gru.risk.workflows.apis.rules import create_rule, get_workflow_rules, update_workflow_rule
from gru.risk.workflows.models.rules import RiskRule, UpdateRuleRequest, UpdateRuleStageRequest, UpdateRuleStatusRequest

class RulesCommands:
    """Commands for managing workflow rules"""
    
    def _format_sub_rules(self, rule_def: Dict[str, Any], indent: int = 2) -> List[str]:
        """Format sub-rules into readable lines"""
        lines = []
        spaces = " " * indent
        
        if "operator" not in rule_def or "sub_rules" not in rule_def:
            return lines
            
        lines.append(f"{spaces}Operator: {rule_def['operator']}")
        
        for sub_rule in rule_def.get("sub_rules", []):
            lines.append(f"{spaces}- Field: {sub_rule.get('field')}")
            lines.append(f"{spaces}  Operation: {sub_rule.get('operator')}")
            if sub_rule.get('value') is not None:
                lines.append(f"{spaces}  Value: {sub_rule['value']}")
            if sub_rule.get('redis_key'):
                lines.append(f"{spaces}  Redis Key: {sub_rule['redis_key']}")
            if sub_rule.get('redis_field'):
                lines.append(f"{spaces}  Redis Field: {sub_rule['redis_field']}")
                
        return lines
    
    def _format_rule(self, rule: RiskRule, detailed: bool = False) -> List[str]:
        """Format a rule into a list of output lines"""
        lines = []
        
        # Basic info
        status_str = f"[{rule.status}]"
        stage_str = f"(Stage: {rule.stage})"
        lines.append(f"{rule.rule_name} {status_str} {stage_str}")
        
        # Add rule definition details if requested
        if detailed:
            lines.extend(self._format_sub_rules(rule.rule_definition))
            lines.append(f"Created: {rule.created_at.strftime('%Y-%m-%d %H:%M:%S UTC')}")
            lines.append(f"Updated: {rule.updated_at.strftime('%Y-%m-%d %H:%M:%S UTC')}")
            lines.append("")
            
        return lines

    def list(
        self,
        workflow_name: str,
        rule_name: Optional[str] = None,
        stage: Optional[str] = None,
        status: Optional[str] = None,
        detailed: bool = False
    ) -> str:
        """
        List rules for a workflow or get details of a specific rule.
        
        Args:
            workflow_name: Name of the workflow
            rule_name: Optional specific rule to retrieve
            stage: Optional stage filter
            status: Optional status filter
            detailed: Whether to show detailed rule information
            
        Returns:
            Formatted string containing rule information
        """
        try:
            auth_token = read_token()
            
            rules = get_workflow_rules(
                auth_token, 
                workflow_name,
                rule_name,
                stage,
                status
            )
            
            if not rules:
                return "No rules found"
            
            output_lines = []
            
            if rule_name:

                output_lines.append(f"Rule details for '{rule_name}':")
            else:

                filters = []
                if stage:
                    filters.append(f"stage={stage}")
                if status:
                    filters.append(f"status={status}")
                    
                header = f"Rules for workflow '{workflow_name}'"
                if filters:
                    header += f" (Filters: {', '.join(filters)})"
                output_lines.append(header)
            
            output_lines.append("")

            for rule in rules:
                output_lines.extend(
                    self._format_rule(rule, detailed=detailed or rule_name is not None)
                )
            
            return "\n".join(output_lines).rstrip()
            
        except ApiError as api_error:
            return f"API Error: {api_error.message}"
        except ValueError as value_error:
            return str(value_error)
        except Exception as e:
            return f"Error: {str(e)}"
        
    def create(self, workflow_name: str, config: str) -> str:
        """
        Create a new rule for a workflow using configuration from a JSON file.
        
        Args:
            workflow_name: Name of the workflow to add the rule to
            config: Path to JSON configuration file
            
        Returns:
            Formatted string with created rule details or error message
        """
        try:
            auth_token = read_token()
            rule_config = read_json_config(config)
                
            rule = create_rule(auth_token, workflow_name, rule_config)
            
            return (
                f"Successfully created rule in workflow '{workflow_name}':\n\n"
                f"{self._format_rule(rule)}"
            )
            
        except ApiError as api_error:
            return f"API Error: {api_error.message}"
        except ValueError as value_error:
            return str(value_error)
        except Exception as e:
            return f"Error: {str(e)}"
    
    def update(
        self,
        workflow_name: str,
        rule_name: str,
        config: Optional[str] = None,
        status: Optional[Literal['ACTIVE', 'INACTIVE', 'DEPRECATED']] = None,
        stage: Optional[Literal['REGISTERED', 'BACKTESTING', 'STAGING', 'PRODEXPERIMENT', 'LIVE']] = None
    ) -> str:
        """
        Update a workflow rule's definition, status, or stage.
        
        Args:
            workflow_name: Name of the workflow
            rule_name: Name of the rule to update
            config: Optional path to rule definition JSON file
            status: Optional new status value
            stage: Optional new stage value
            
        Returns:
            Formatted string showing updated rule details
            
        Note: Exactly one of config, status, or stage must be provided
        """
        try:
            # Validate input combination
            provided_options = sum(bool(x) for x in [config, status, stage])
            if provided_options != 1:
                return "Error: Exactly one of --config, --status, or --stage must be provided"
            
            auth_token = read_token()
            
            if config:
                config_data = read_json_config(config)
                request = UpdateRuleRequest.model_validate(config_data)
                update_type = 'definition'
                request_data = request
            elif status:
                request = UpdateRuleStatusRequest(new_status=status)
                update_type = 'status'
                request_data = request
            elif stage:  
                request = UpdateRuleStageRequest(stage=stage)
                update_type = 'stage'
                request_data = request
            
            updated_rule = update_workflow_rule(
                auth_token,
                workflow_name,
                rule_name,
                update_type,
                request_data
            )
            
            update_type_str = 'definition' if config else 'status' if status else 'stage'
            return "\n".join([
                f"Successfully updated rule {update_type_str}:",
                "",
                *self._format_rule(updated_rule, detailed=True)
            ])
            
        except ApiError as api_error:
            return f"API Error: {api_error.message}"
        except ValueError as value_error:
            return f"Validation error: {str(value_error)}"
        except Exception as e:
            return f"Error: {str(e)}"




import os
from typing import Any
from celery import Celery
from pydantic import BaseModel

broker = os.getenv("BROKER_URL")

task_server = Celery("task_server", broker=broker)


class TaskMetadata(BaseModel):
    prompt_id: str
    tool_call_id: str
    agent_name: str
    agent_callback_url: str


def submit_task(
    task_type: str, metadata: TaskMetadata, task_attributes: dict[str, Any]
):
    task_metadata = metadata.model_dump()
    task_server.send_task(
        "src.tasks.execute", args=[task_type, task_metadata, task_attributes]
    )



"""
ML Server Starter

This script starts a machine learning server based on the configuration provided
in a YAML file. It dynamically loads a Flask application from the specified entry
point and launches a web server, supporting different types such as Gunicorn.

Usage:
    python server.py --config-file <path_to_config>

Options:
    --config-file (str): Path to the configuration YAML file.
"""
import click
import os
from gru.ml_serving.web_frameworks.flask_adapter import FlaskAdapter
from gru.ml_serving.servers.gunicorn import GunicornServer
from .ml_serving_config_reader import MLServingConfigReader

GUNICORN_TYPE = 'gunicorn'

@click.command()
@click.option('--config-file', type=click.Path(exists=True), help='Path to the config YAML file')
def serve(config_file):
    """
    Start the server based on the configuration provided in the YAML file.

    Parameters:
    - config_file (str): Path to the config YAML file.
    """
    try:
        config = MLServingConfigReader(config_file)
        for key, value in config.get_config('env_vars').items():
            click.echo(f'Setting environment variable: {key}={value}')
            os.environ[key] = value

        application = load_application(config)
        start_web_server(config, application)

    except Exception as e:
        click.echo(f"Error: {str(e)}")

def load_application(config: MLServingConfigReader):
    """
    Load the application module dynamically.

    Parameters:
    - entry_point (str): Path to the entry point module.

    Returns:
    - WebServerAdapter: The application instance.
    """
    try:
        serving_layer = config.get_config('serving_layer')
        if serving_layer == 'flask':
            framework_adapter = FlaskAdapter(config)
        else:
            raise Exception(f"Serving layer {serving_layer} is not supported yet")

        print(f"Loading {serving_layer} application from entrypoint: {config.get_config('entrypoint')}")
        return framework_adapter.load_application()

    except Exception as e:
        raise Exception(f"Error loading {serving_layer} application: {str(e)}")

def start_web_server(config: MLServingConfigReader, application):
    """
    Start the web server based on the specified configuration.

    Parameters:
    - config (MLServingConfigReader): The server configuration reader.
    - application: The Flask application instance.
    """
    try:
        webserver_config = config.get_config('webserver')
        type = webserver_config.get('type')
        config = webserver_config.get('config')
        if type == GUNICORN_TYPE:
            server = GunicornServer(application, config)
        else:
            raise Exception(f"The web server of type '{webserver_config.get('type')}' is not supported yet.")

        print(f"String webserver {type} with configuration: {config}")
        server.start()

    except Exception as e:
        raise RuntimeError(f"Error starting web server: {str(e)}")

if __name__ == '__main__':
    serve()



"""
In Machine Leaning, a feature is an individual measurable property or characteristic of a phenomenon. Features are used for

 - model training
 - inference/model prediction
 - exploratory analyses
 - Features can be defined using the FeatureLogic class 1. Other operational inputs such as feature_start_time, schedule etc are used to convert the a Feature Definition to a DAG.

 Function:
    - print_feature
    - register_feature
    - deploy_feature
"""
import datetime
from typing import Any, Dict, List, Optional, Union
from datetime import datetime
from gru.features.ml_feature import MLFeature
from gru.utils.data_types import DataType
from gru.features.custom_feature_logic import CustomFeatureLogic
from gru.features.feature_logic import FeatureLogic
from gru.utils.entity_type import EntityType
from gru.utils.processing_engine import ProcessingEngine
from gru.features.processing_engine_configs import BatchProcessingEngineConfigs


class RawFeature(MLFeature):
    """
    A raw feature is a computed column, calculated on top of a field by performing an aggregation funciton for a given timeperiod of data.

    Parameters
    ----------
    self: str
        Python self object
    name : str
        The name of the feature.
    source_name : List(str
        The name of the data source from which the feature is calculated.
    sink_name : List(str
        The name of the destination where the feature is stored.
    online : bool
        If True, the feature will be persisted to the Online Feature Store. Default is False.
    offline : bool
        If True, the feature will be persisted to the Offline Feature Store. Default is True.
    schedule : str
        Cron schedule for the feature generation. Ref: https://crontab.guru/
    entity : List[str]
        List of entities associated with the feature.
    feature_logic : FeatureLogic
        FeatureLogic object defining the logic of the feature.
    description : str
        Description of the feature. Default is an empty string.
    data_type : DataTypes
        Data type of the feature
    owners : List[str]
        List of owner names. Default is an empty list.
    active : bool
        If True, the feature will start running as soon as it's deployed. Default is True.
    timeout : str
        Maximum time for feature calculation before sending an alert. (Not supported as part of MVP)
    k8s_configs : Dict
        Kubernetes configurations for feature processing.
    processing_engine_configs : Dict
        Processing engine configurations.
    start_time : datetime
        Start time of the feature. Default is the current datetime.
    creation_time : datetime
        Timestamp of feature creation. Default is the current datetime.


    Returns
    -------
    None

    Notes
    -----
    Use the class to create Raw features for both batch and streaming sources

    Examples:
    >>> create_raw_feature = RawFeature(
    ...     name="va_testing_rows_4",
    ...     description="test",
    ...     data_type=DataType.FLOAT,
    ...     source_name=["svt_half_hour2"],
    ...     sink_name=["sink_test"],
    ...     owners=["vanshika@yugen.ai"],
    ...     entity=["test"],
    ...     feature_logic=FeatureLogic(
    ...         field=["ad_id"],
    ...         filter="ad_id is NOT NULL",
    ...         transform=SlidingWindowAggregation(
    ...             function="avg",
    ...             partition_by="provider",
    ...             order_by="cpi",
    ...             # rangeBetween= {"frame_start": 1, "frame_end": 6},
    ...             rowsBetween={"frame_start": 1, "frame_end": 2},
    ...         ),
    ...         time_window="3d",
    ...         groupby_keys=["project_id"],
    ...         timestamp_field="time",
    ...         timestamp_format="yyyy-MM-dd HH:mm:ssXXX",
    ...     ),
    ...     processing_engine=ProcessingEngine.PYSPARK_K8S,
    ...     processing_engine_configs=ProcessingEngineConfigs(config_path="/user_specific/path/config.yaml"),
    ...     online=True,
    ...     offline=True,
    ...     schedule="0 0 * * *",
    ...     active=True,
    ...     start_time=datetime(2023, 3, 29, 0, 0, 0),
    ... )
    """

    def __init__(
        self,
        name: str,
        data_type: DataType,
        data_sources: List[str],
        offline: bool,
        staging_sink: List[str],
        schedule: str,
        entity: List[str],
        feature_logic: Union[FeatureLogic, CustomFeatureLogic],
        start_time: datetime,
        online: bool = False,
        online_sink: List[str] = [],
        description: str = "",
        owners: List[str] = [],
        active: bool = False,
        timeout: str = "5m",
        k8s_configs: Dict = {},
        read_option_configs: Optional[List[Dict[str, Any]]] = None,
        staging_sink_write_option_configs: Optional[Dict[str, Any]] = None,
        online_sink_write_option_configs: Optional[Dict[str, Any]] = None,
        processing_engine: Optional[ProcessingEngine] = ProcessingEngine.PYSPARK_K8S,
        processing_engine_configs: Optional[BatchProcessingEngineConfigs] = None,
        creation_time: datetime = datetime.now(),
        tags: Dict[str, Any] = {},
    ):
        airflow_configs = {"active": active, "timeout": timeout, "schedule": schedule}
        sink = {"staging": staging_sink, "online": online_sink}
        super().__init__(
            name=name,
            type=EntityType.RAW_FEATURE.value,
            data_type=data_type,
            online=online,
            offline=offline,
            entity=entity,
            data_sources=data_sources,
            sink=sink,
            feature_logic=feature_logic,
            description=description,
            owners=owners,
            start_time=start_time,
            k8s_configs=k8s_configs,
            processing_engine=processing_engine,
            processing_engine_configs=processing_engine_configs
            or BatchProcessingEngineConfigs(),
            airflow_configs=airflow_configs,
            creation_time=creation_time,
            tags=tags,
            read_option_configs=read_option_configs,
            staging_sink_write_option_configs=staging_sink_write_option_configs,
            online_sink_write_option_configs=online_sink_write_option_configs,
        )

    def to_json(self) -> Dict:
        return {
            "name": self.name,
            "type": self.type,
            "data_type": self.data_type.value,
            "data_sources": self.data_sources,
            "owners": self.owners,
            "description": self.description,
            "start_time": str(self.start_time),
            "online": self.online,
            "offline": self.offline,
            "entity": self.entity,
            "feature_logic": self.feature_logic.to_json(),
            "k8s_configs": self.k8s_configs,
            "processing_engine": self.processing_engine.value,
            "processing_engine_configs": self.processing_engine_configs.to_json(),
            "sink": self.sink,
            "creation_time": str(self.creation_time),
            "airflow_configs": self.airflow_configs,
            "created_at": str(self.creation_time),
            "metadata" : self.tags,
            "read_option_configs": self.read_option_configs,
            "staging_sink_write_option_configs": self.staging_sink_write_option_configs,
            "online_sink_write_option_configs": self.online_sink_write_option_configs,
        }

    def to_register_json(self) -> Dict:
        return self.to_json()



QUERY_ANALYZER_SYSTEM_PROMPT = """
You are an expert SQL analyst. Analyze natural language queries to identify key entities
and business domains. Focus on SQL-relevant entities like tables, columns, and business concepts.
"""

QUERY_ANALYZER_USER_PROMPT_TEMPLATE = """
Analyze this query and identify entities and business domains:
Query: {query}

Return only the entities and domains in this format:
Entities: [list of entities]
Domains: [list of business domains]
"""



from abc import ABC, abstractmethod
from typing import List, Dict, Any, Optional
from pydantic import BaseModel


class VectorDBConfig(BaseModel):
    host: str
    port: int
    user: str
    password: str
    connection_timeout: int
    max_retries: int

class CreateCollectionResponse(BaseModel):
    status: str
    message: str
    collection_name: str

class AddToCollectionResponse(BaseModel):
    status: str
    message: str
    ids: List[str]

class UpdateCollectionResponse(BaseModel):
    status: str
    message: str
    id: str
    updated_count: Optional[int] = None

class DeleteFromCollectionResponse(BaseModel):
    status: str
    message: str

class SimilaritySearchResponse(BaseModel):
    status: str
    message: str
    results: List[Dict[str, Any]]

class ListCollectionsResponse(BaseModel):
    status: str
    message: str
    collections: List[str]

class GetCollectionInfoResponse(BaseModel):
    status: str
    message: str
    collection_info: Dict[str, Any]

class DropCollectionResponse(BaseModel):
    status: str
    message: str

class VectorDBClient(ABC):
    @abstractmethod
    def connect(self, config: VectorDBConfig) -> None:
        """Initialize connection to vector database"""
        pass

    @abstractmethod
    def create_collection_sync(
        self,
        collection_name: str,
        schema: Dict[str, Any],
        replace_existing: bool = False
    ) -> CreateCollectionResponse:
        """Create a new collection with specified schema synchronously"""
        pass

    @abstractmethod
    async def create_collection(
        self,
        collection_name: str,
        schema: Dict[str, Any],
        replace_existing: bool = False
    ) -> CreateCollectionResponse:
        """Create a new collection with specified schema asynchronously"""
        pass

    @abstractmethod
    async def add_to_collection(
        self, 
        collection_name: str, 
        embeddings: List[List[float]], 
        data: List[Dict[str, Any]]
    ) -> AddToCollectionResponse:
        """Add documents with embeddings to collection"""
        pass

    @abstractmethod
    async def update_collection(
        self, 
        collection_name: str, 
        memory_id: str, 
        embedding: List[float],
        data: Optional[Dict[str, Any]] = None
    ) -> UpdateCollectionResponse:
        """Update document in collection"""
        pass

    @abstractmethod
    async def delete_from_collection(
        self, 
        collection_name: str, 
        memory_id: str
    ) -> DeleteFromCollectionResponse:
        """Delete document from collection"""
        pass

    @abstractmethod
    async def similarity_search(
        self,
        collection_name: str,
        query_vector: List[float],
        anns_field: Optional[str] = None,
        top_k: int = 5,
        search_params: Optional[Dict[str, Any]] = None,
        output_fields: Optional[List[str]] = None,
    ) -> SimilaritySearchResponse:
        """Search for similar vectors in collection"""
        pass

    @abstractmethod
    async def filtered_search(
        self,
        collection_name: str,
        filter_expr: str,
        limit: int = 5,
        output_fields: Optional[List[str]] = None,
    ) -> SimilaritySearchResponse:
        """Performs a filtered search based on additional query parameters."""
        pass

    @abstractmethod
    async def list_collections(self) -> ListCollectionsResponse:
        """List all available collections"""
        pass

    @abstractmethod
    async def get_collection_info(
        self, 
        collection_name: str
    ) -> GetCollectionInfoResponse:
        """Get collection metadata and statistics"""
        pass

    @abstractmethod
    async def drop_collection(
        self, 
        collection_name: str
    ) -> DropCollectionResponse:
        """Delete an entire collection"""
        pass


from pydantic import BaseModel, validator
from typing import Dict, Any

class BuiltInTransformation(BaseModel):
    """
    Pydantic model for validating built-in Spark transformations configurations.

    This model validates configurations for Spark's built-in transformations like
    filter, selectExpr, and repartition.

    Attributes:
        type (str): Type of built-in transformation ('filter', 'selectExpr', 'repartition')
        additional_params (Dict[str, Any]): Parameters required for the transformation

    Examples:
        >>> # Valid filter transformation
        >>> config = BuiltInTransformation(
        ...     type="filter",
        ...     additional_params={"condition": "col('age') > 18"}
        ... )

        >>> # Valid selectExpr transformation
        >>> config = BuiltInTransformation(
        ...     type="selectExpr",
        ...     additional_params={"expression": "upper(name) as uppercase_name"}
        ... )

        >>> # Invalid configuration (will raise ValueError)
        >>> config = BuiltInTransformation(
        ...     type="filter",
        ...     additional_params={}  # Missing required 'condition' parameter
        ... )
    """

    type: str
    additional_params: Dict[str, Any]

    @validator('additional_params')
    def validate_params(cls, v, values):
        type = values.get('type')
        required_params = {
            'filter': ['condition'],
            'selectExpr': ['expression'],
            'repartition': ['numPartitions'],
        }

        if type not in required_params:
            raise ValueError(f"Unsupported transformation type: {type}")

        for param in required_params[type]:
            if param not in v:
                raise ValueError(f"{type.capitalize()} transformation must have a '{param}' parameter")

        return v

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": self.type,
            "additional_params": self.additional_params
        }

class LambdaTransformation(BaseModel):
    """
    Pydantic model for validating Lambda UDF (User Defined Function) transformations.

    This model validates configurations for map and flatMap transformations that use
    lambda functions as UDFs in Spark.

    Attributes:
        type (str): Type of transformation ('map' or 'flatMap')
        function (str): Lambda function as string

    Examples:
        >>> # Valid map transformation
        >>> config = LambdaTransformation(
        ...     type="map",
        ...     function="lambda x: x * 2"
        ... )

        >>> # Valid flatMap transformation
        >>> config = LambdaTransformation(
        ...     type="flatMap",
        ...     function="lambda text: text.split()"
        ... )

        >>> # Invalid function (will raise ValueError)
        >>> config = LambdaTransformation(
        ...     type="map",
        ...     function="def multiply(x): return x * 2"  # Not a lambda function
        ... )
    """

    type: str
    function: str

    @validator('type')
    def validate_type(cls, v):
        allowed_types = ['map', 'flatMap']
        if v not in allowed_types:
            raise ValueError(f"Unsupported transformation type: {v}. Allowed types are {allowed_types}")
        return v

    @validator('function')
    def validate_function(cls, v, values):
        type = values.get('type')
        if not v.startswith('lambda'):
            raise ValueError(f"Function must be a lambda expression")

        # Split the lambda function to get parameters
        params = v.split(':')[0].strip().lstrip('lambda').strip().split(',')

        if type == 'map':
            if len(params) < 1:
                raise ValueError("Map function must have at least one parameter")
        elif type == 'flatMap':
            if len(params) < 1:
                raise ValueError("FlatMap function must have at least one parameter")
        return v

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": self.type,
            "function": self.function,
        }



from typing import Optional, Type
from langchain_core.tools import BaseTool
from pydantic import BaseModel, Field
from langchain_core.callbacks import CallbackManagerForToolRun
from gru.agents.tools.core.code_loader.github import GitHubLoader
from gru.agents.tools.core.code_analyzer.argument import ArgumentAnalyzer


class PythonArgumentsToolInput(BaseModel):
    repository: str = Field(description="repository containing the python file")
    file_path: str = Field(description="path of the python file in the repository")


class PythonArgumentsTool(BaseTool):
    name: str = "get_python_file_arguments"
    description: str = (
        "Use this to determine the command line arguments that are required to execute a python file present in a repository"
    )
    args_schema: Type[BaseModel] = PythonArgumentsToolInput
    return_direct: bool = True

    class Config:
        extra = "allow"

    def __init__(self, token: str):
        super().__init__(token=token)
        self.token = token
        self.code_loader = GitHubLoader(token)
        self.analyzer = ArgumentAnalyzer()

    def _run(self, *args, **kwargs):
        return super()._run(*args, **kwargs)
    
    async def _arun(
        self,
        repository: str,
        file_path: str,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ):
        
        source = self.code_loader.retrieve_code(repository, file_path)
        
        
        arg_details = self.analyzer.analyze(source)

        return arg_details



import os
import threading

from psycopg_pool import AsyncConnectionPool


class PostgresAsyncConnectionPool:
    _instance = None
    _lock = threading.Lock()

    def __new__(cls):
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super(PostgresAsyncConnectionPool, cls).__new__(cls)
                    cls._instance._initialize_pool()
        return cls._instance

    def _initialize_pool(self):
        db_host = os.getenv("CHECKPOINT_DB_HOST")
        db_name = os.getenv("DB_NAME", "postgres")
        db_user = os.getenv("CHECKPOINT_DB_USER", "postgres")
        db_password = os.getenv("CHECKPOINT_DB_PASSWORD")
        db_port = os.getenv("CHECKPOINT_DB_PORT", "5432")
        db_max_connections = int(os.getenv("DB_MAX_CONNECTIONS", "10"))

        db_uri = f"postgresql://{db_user}:{db_password}@{db_host}:{db_port}/{db_name}?sslmode=disable"

        connection_pool = AsyncConnectionPool(
            conninfo=db_uri,
            max_size=db_max_connections,
            kwargs={
                "autocommit": True,
                "prepare_threshold": 0,
                "keepalives": 1,
                "keepalives_idle": 30,
                "keepalives_interval": 10,
                "keepalives_count": 5,
            },
        )

        self.pool = connection_pool

    def get(self):
        return self.pool



import pytest, os
from gru.features.raw_feature import RawFeature
from gru.utils.data_types import DataType
from gru.features.feature_logic import FeatureLogic
from gru.utils.processing_engine import ProcessingEngine
from gru.utils.read_options import KafkaReadOptions
from gru.features.processing_engine_configs import BatchProcessingEngineConfigs
from datetime import datetime
from gru.features.aggregations import SlidingWindowAggregation


@pytest.fixture
def raw_feature_fixture():
    current_dir = os.path.dirname(os.path.abspath(__file__))
    configs_path = os.path.join(
        current_dir, "data", "test_default_processing_config_batch.yaml"
    )
    raw_feature = RawFeature(
        name="testing_raw_feature",
        description="test",
        data_type=DataType.FLOAT,
        staging_sink=["foo"],
        staging_sink_write_option_configs={
            "foo": {
                "bucket": "internal-ml-demos",
            },
        },
        online_sink=["bar"],
        online_sink_write_option_configs={
            "bar": {"file_type_properties": {"type": "PARQUET", "mergeSchema": False}},
        },
        data_sources=["analytics_data", "survey_telemetry_data"],
        read_option_configs=[
            {
                "name": "kafka_streaming_source",
                "options": KafkaReadOptions(
                    watermark_delay_threshold="10 seconds",
                    processing_time="120 seconds",
                    starting_offsets="earliest",
                    include_headers=True,
                ).to_json(),
            }
        ],
        owners=["foo2@yugen.ai"],
        entity=["quux"],
        feature_logic=FeatureLogic(
            field=["corge"],
            filter="grault",
            transform=SlidingWindowAggregation(
                function="garply",
                partition_by="waldo",
                order_by="fred",
                rowsBetween={"frame_start": 1, "frame_end": 2},
            ),
            time_window="3d",
            groupby_keys=["plugh"],
            timestamp_field="time",
            timestamp_format="yyyy-MM-dd HH:mm:ssXXX",
        ),
        processing_engine=ProcessingEngine.PYSPARK_K8S,
        processing_engine_configs=BatchProcessingEngineConfigs(config_path=configs_path),
        online=True,
        offline=True,
        schedule="0 0 * * *",
        active=True,
        start_time=datetime(2023, 12, 5, 0, 0, 0),
        creation_time="2023-12-05 18:06:57.175721",
    )
    return raw_feature


def test_to_json(raw_feature_fixture):
    """
    Takes a custom raw fixture to test the default parameters

    Parameters
    ----------
    raw_feature_fixture : fixture object

    Returns
    -------
    pytest true/false
    """
    expected_result = {
        "name": "testing_raw_feature",
        "type": "raw_feature",
        "data_type": "float",
        "data_sources": ["analytics_data", "survey_telemetry_data"],
        "owners": ["foo2@yugen.ai"],
        "description": "test",
        "start_time": "2023-12-05 00:00:00",
        "online": True,
        "offline": True,
        "entity": ["quux"],
        "feature_logic": {
            "is_feature_logic_custom": False,
            "name": "",
            "transform": {
                "function": "garply",
                "partition_by": "waldo",
                "order_by": "fred",
                "rowsBetween": {"frame_start": 1, "frame_end": 2},
                "function_type": "SlidingWindowAggregation",
            },
            "time_window": "3d",
            "groupby_keys": ["plugh"],
            "field": ["corge"],
            "filter": "grault",
            "timestamp_field": "time",
            "timestamp_format": "yyyy-MM-dd HH:mm:ssXXX",
        },
        "k8s_configs": {},
        "processing_engine": "pyspark_k8s",
        "processing_engine_configs": {
            "deploy-mode": "foo1",
            "num-executors": 10,
            "spark.driver.cores": 1,
            "spark.executor.cores": 1,
            "spark.executor.instances": 1,
            "spark.hadoop.fs.s3a.aws.credentials.provider": "foo3",
            "spark.kubernetes.authenticate.driver.serviceAccountName": "bar2",
            "spark.kubernetes.container.image.pullSecrets": "bar3",
            "spark.kubernetes.driver.limit.cores": "bar4",
            "spark.kubernetes.driver.memory": "bar1",
            "spark.kubernetes.executor.limit.cores": "foo5",
            "spark.kubernetes.executor.memory": "foo2",
            "spark.kubernetes.file.upload.path": "/path/to/config/file",
            "spark.kubernetes.namespace": "foo4",
        },
        "sink": {"staging": ["foo"], "online": ["bar"]},
        "read_option_configs": [
            {
                "name": "kafka_streaming_source",
                "options": {
                    "delimiter": "|",
                    "fail_on_data_loss": False,
                    "include_headers": True,
                    "infer_schema": True,
                    "processing_time": "120 seconds",
                    "starting_offsets": "earliest",
                    "starting_offsets_by_timestamp": None,
                    "starting_timestamp": None,
                    "watermark_delay_threshold": "10 " "seconds",
                },
            }
        ],
        "online_sink_write_option_configs": {
            "bar": {"file_type_properties": {"type": "PARQUET", "mergeSchema": False}}
        },
        "staging_sink_write_option_configs": {"foo": {"bucket": "internal-ml-demos"}},
        "creation_time": "2023-12-05 18:06:57.175721",
        "airflow_configs": {"active": True, "timeout": "5m", "schedule": "0 0 * * *"},
        "created_at": "2023-12-05 18:06:57.175721",
        "metadata": {},
    }

    assert raw_feature_fixture.to_json() == expected_result



from typing import Optional, Type, List, Dict, Any
from pydantic import BaseModel
from langchain_core.tools import BaseTool
from langchain_core.callbacks import CallbackManagerForToolRun
from gru.agents.tools.core.services.text_to_sql import TextToSQLService
from gru.agents.tools.text_to_sql.models import TextToSQLToolInput


class TextToSQLTool(BaseTool):
    name: str = "text_to_sql_converter"
    description: str = "Use this tool to convert natural language queries into SQL."
    args_schema: Type[BaseModel] = TextToSQLToolInput
    return_direct: bool = True

    class Config:
        extra = "allow"

    def __init__(self, service: TextToSQLService):
        super().__init__()
        self.service = service

    def _run(self, *args, **kwargs):
        return super()._run(*args, **kwargs)

    async def _arun(
        self,
        query: str,
        table_info: Optional[List[Dict[str, Any]]] = None,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> str:
        try:
            sql = await self.service.convert_to_sql(query=query, table_info=table_info)
            return sql
        except Exception as e:
            if run_manager:
                run_manager.on_tool_error(e)
            return f"Error converting text to SQL: {str(e)}"



from gru.agents.tools.langgraph.git_repo_contents import GitRepoContentRetriever
from gru.agents.tools.langgraph.python_arguments import PythonArgumentsToolInput
from gru.agents.tools.langgraph.python_code_runner import PythonCodeRunner
from gru.agents.tools.langgraph.python_run_status import PythonRunStatusChecker
from gru.agents.tools.langgraph.sql_runner import CansoSQLRunnerTool
from gru.agents.tools.langgraph.data_quality_tool import DataQualityTool
from gru.agents.tools.langgraph.memory_tools import MemoryRetrievalTool



from enum import Enum

class SinkFamily(Enum):
    """
    Enum class representing various sink families
    """

    REDIS = "redis"
    OBJECT_STORAGE = "object_storage"

    @classmethod
    def get_key(cls, value: str):
        """
        Retrieve the key (name) of a sink family based on its string value.

        Args:
            value: The string value representing a sink family.
        
        Returns:
            str: The key (name) of the sink family.

        Raises:
            ValueError: If no key is found for the given value.

        Example:
        >>> SinkFamily.get_key("redis")
        'REDIS'
        """
        for item in cls:
            if item.value == value:
                return item.name
        raise ValueError(f"sink family \"{value}\" isn't supported yet.")


from enum import StrEnum

from gru.agents.checkpoint.postgres import PostgresAsyncConnectionPool


CREATE_TABLE_QUERY = """
    CREATE TABLE IF NOT EXISTS canso_task_results (
        agent_name varchar(64) NOT NULL,
        prompt_id varchar(64) NOT NULL,
        task_type varchar(64) NOT NULL,
        run_id varchar(64) NOT NULL,
        status varchar(32) NOT NULL,
        result TEXT,
        PRIMARY KEY (agent_name, prompt_id, task_type, run_id)
    )
"""

INSERT_QUERY = """
    INSERT INTO canso_task_results (agent_name, prompt_id, task_type, run_id, status, result)
    VALUES (%s, %s, %s, %s, %s, %s)
    ON CONFLICT (agent_name, prompt_id, task_type, run_id) 
    DO UPDATE SET 
        status = EXCLUDED.status,
        result = EXCLUDED.result;
"""

select_query = """
    SELECT status, result 
    FROM canso_task_results 
    WHERE agent_name = %s AND prompt_id = %s AND task_type = %s AND run_id = %s
"""


class TaskStatus(StrEnum):
    PROCESSING = "PROCESSING"
    COMPLETED = "COMPLETED"


class TaskResultsRepository:

    async def setup(self):
        connection_pool = PostgresAsyncConnectionPool().get()
        async with connection_pool.connection() as conn:
            async with conn.cursor() as cur:
                await cur.execute(CREATE_TABLE_QUERY)

    async def update(
        self,
        agent_name: str,
        prompt_id: str,
        task_type: str,
        run_id: str,
        status: TaskStatus,
        result: str = None,
    ):
        connection_pool = PostgresAsyncConnectionPool().get()
        async with connection_pool.connection() as conn:
            async with conn.cursor() as cur:
                await cur.execute(
                    INSERT_QUERY,
                    (agent_name, prompt_id, task_type, run_id, status, result),
                )

    async def get_result(
        self, agent_name: str, prompt_id: str, task_type: str, run_id: str
    ):
        connection_pool = PostgresAsyncConnectionPool().get()
        async with connection_pool.connection() as conn:
            async with conn.cursor() as cur:
                await cur.execute(
                    select_query, (agent_name, prompt_id, task_type, run_id)
                )
                row = await cur.fetchone()

                if row is None:
                    return None

                return row[0], row[1]
